#External:
import multiprocessing
import datetime
#import sys
#from StringIO import StringIO
import netCDF4
import requests
import requests_cache

#Internal:
from . import netcdf_utils, requests_sessions, retrieval_utils
from .certificates import certificates
from .remote_netcdf import remote_netcdf

def start_download_processes(options,q_manager,previous_processes=dict()):
    remote_netcdf_kwargs=dict()
    if 'download_cache' in dir(options) and options.download_cache:
        remote_netcdf_kwargs['cache']=options.download_cache.split(',')[0]
        if len(options.download_cache.split(','))>1:
            remote_netcdf_kwargs['expire_after']=datetime.timedelta(hours=float(options.download_cache.split(',')[1]))

    #Add credentials:
    remote_netcdf_kwargs.update({opt: getattr(options,opt) for opt in ['openid','username','password','use_certificates'
                                                                     ] if opt in dir(options)})
    #This allows time variables with different names:
    if 'time_var' in dir(options) and options.time_var:
        time_var=options.time_var
    else:
        time_var='time'

    #Start processes for download. Can be run iteratively for an update.
    processes=previous_processes
    if not ('serial' in dir(options) and options.serial):
        processes=start_download_processes_no_serial(q_manager,options.num_dl,processes,time_var=time_var,
                                                                                        remote_netcdf_kwargs=remote_netcdf_kwargs)
    return processes

def start_download_processes_no_serial(q_manager,num_dl,processes,time_var='time',remote_netcdf_kwargs=dict()):
    for data_node in q_manager.queues.keys():
        for simultaneous_proc in range(num_dl):
            process_name=data_node+'-'+str(simultaneous_proc)
            if not process_name in processes.keys():
                processes[process_name]=multiprocessing.Process(target=worker_retrieve, 
                                                name=process_name,
                                                args=(q_manager,data_node,time_var,remote_netcdf_kwargs))
                processes[process_name].start()
    return processes

def worker_retrieve(q_manager,data_node,time_var,remote_netcdf_kwargs):
    if ( 'session' in dir(q_manager) and
        (isinstance(q_manager.session,requests.Session) or
            isinstance(q_manager.session,requests_cache.core.CachedSession)
            )):
        session=q_manager.session
    else:
        #Create one session per worker:
        session=requests_sessions.create_single_session(**remote_netcdf_kwargs)

    #Loop indefinitely. Worker will be terminated by main process.
    while True:
        item = q_manager.queues.get(data_node)
        if item=='STOP': break
        try:
            thread_id=item[0]
            trial=item[1]
            path_to_retrieve=item[2]
            file_type=item[3]
            remote_data=remote_netcdf.remote_netCDF(path_to_retrieve,file_type,session=session,
                                                                               time_var=time_var,
                                                                               **remote_netcdf_kwargs)

            var_to_retrieve=item[4]
            pointer_var=item[5]
            result=remote_data.download(var_to_retrieve,pointer_var,download_kwargs=item[-1])
            q_manager.put_for_thread_id(thread_id,(file_type,result))
        except Exception:
            if trial==3:
                print('Download failed with arguments ',item)
                raise
                q_manager.put_for_thread_id(thread_id,(file_type,'FAIL'))
            else:
                #Put back in the queue. Do not raise. Simply put back in the queue so that failure
                #cannnot occur while working downloads work:
                item_new=(trial+1,path_to_retrieve,file_type, var_to_retrieve,pointer_var,item[-1])
                q_manager.put_again_to_data_node_from_thread_id(thread_id,data_node,item_new)
    return

def worker_exit(q_manager,data_node_list,queues_size,start_time,renewal_time,output,options):
    failed=False
    while True:
        item = q_manager.get_for_thread_id()
        if item=='STOP': break
        renewal_time,failed=progress_report(item[0],item[1],q_manager,data_node_list,queues_size,start_time,renewal_time,failed,output,options)
    return renewal_time, failed

def launch_download(output,data_node_list,q_manager,options):
    remote_netcdf_kwargs=dict()
    if 'download_cache' in dir(options) and options.download_cache:
        remote_netcdf_kwargs['cache']=options.download_cache.split(',')[0]
        if len(options.download_cache.split(','))>1:
            remote_netcdf_kwargs['expire_after']=datetime.timedelta(hours=float(options.download_cache.split(',')[1]))

    start_time = datetime.datetime.now()
    renewal_time = start_time
    queues_size=dict()
    if 'silent' in dir(options) and not options.silent:
        for data_node in data_node_list:
            queues_size[data_node]=q_manager.queues.qsize(data_node)
        string_to_print=['0'.zfill(len(str(queues_size[data_node])))+'/'+str(queues_size[data_node])+' paths from "'+data_node+'"' for
                            data_node in data_node_list if queues_size[data_node]>0 ]
        if len(string_to_print)>0:
            print('Remaining retrieval from data nodes:')
            print(' | '.join(string_to_print))
            print('Progress: ')

    if 'serial' in dir(options) and options.serial:
        for data_node in data_node_list:
            q_manager.queues.put(data_node,'STOP')
            worker_retrieve(q_manager,data_node,remote_netcdf_kwargs)
            renewal_time, failed=worker_exit(q_manager,data_node_list,queues_size,start_time,renewal_time,output,options)
    else:
        renewal_time, failed=worker_exit(q_manager,data_node_list,queues_size,start_time,renewal_time,output,options)

    if failed:
        raise Exception('Retrieval failed')
                
    if ( 'silent' in dir(options) and not options.silent and
         len(string_to_print)>0):
        print
        print('Done!')
    return output

def progress_report(file_type,result,q_manager,data_node_list,queues_size,start_time,renewal_time,failed,output,options):
    elapsed_time = datetime.datetime.now() - start_time
    renewal_elapsed_time=datetime.datetime.now() - renewal_time

    if file_type=='HTTPServer':
        if not result=='FAIL':
            if 'silent' in dir(options) and not options.silent:
                #print '\t', queues['end'].get()
                if result!=None:
                    print('\t' + result)
                    print(str(elapsed_time))
        else:
            failed=True
    else:
        if not result=='FAIL':
            assign_tree(output,*result)
            output.sync()
            if 'silent' in dir(options) and not options.silent:
                string_to_print=[str(queues_size[data_node]-q_manager.queues.qsize(data_node)).zfill(len(str(queues_size[data_node])))+
                                 '/'+str(queues_size[data_node]) for
                                    data_node in data_node_list if queues_size[data_node]>0]
                print(str(elapsed_time)+', '+' | '.join(string_to_print)+'\r'),
        else:
            failed=True

    #Maintain certificates:
    if ('username' in dir(options) and 
        options.username!=None and
        options.password!=None and
        options.use_certificates and
        renewal_elapsed_time > datetime.timedelta(hours=1)):
        #Reactivate certificates:
        certificates.retrieve_certificates(options.username,'ceda',user_pass=options.password)
        renewal_time=datetime.datetime.now()
    return renewal_time, failed

def assign_tree(output,val,sort_table,tree):
    if len(tree)>1:
        if tree[0]!='':
            assign_tree(output.groups[tree[0]],val,sort_table,tree[1:])
        else:
            assign_tree(output,val,sort_table,tree[1:])
    else:
        output.variables[tree[0]][sort_table,...]=val
    return



#External:
import numpy as np
import netCDF4
import os
import copy
import datetime
import tempfile

#Internal:
from ..remote_netcdf import remote_netcdf,http_netcdf
from .. import netcdf_utils, indices_utils, queues_manager

file_unique_id_list=['checksum_type','checksum','tracking_id']

def _nonzeroprod(x):
    return np.prod(np.asarray(x)[np.nonzero(x)])

class read_netCDF_pointers:
    def __init__(self,data_root,
                    download_all_files=False,
                    download_all_opendap=False,
                    min_year=None,
                    year=None,
                    month=None,
                    day=None,
                    hour=None,
                    previous=0,
                    next=0,
                    requested_time_restriction=[],
                    time_var='time',
                    q_manager=None,
                    session=None,
                    remote_netcdf_kwargs={}):
        self.data_root=data_root
        self.q_manager=q_manager
        self.remote_netcdf_kwargs=remote_netcdf_kwargs
        self.session=session

        self.download_all_files=download_all_files
        self.download_all_opendap=download_all_opendap

        self.time_var = netcdf_utils.find_time_var(self.data_root,time_var=time_var)
        if self.time_var!=None and len(self.data_root.variables[self.time_var])>0:
            #Then find time axis, time restriction and which variables to retrieve:
            self.date_axis = netcdf_utils.get_date_axis(self.data_root,self.time_var)
            self.time_axis = self.data_root.variables[self.time_var][:]
            if len(requested_time_restriction) == len(self.date_axis):
                self.time_restriction = np.array(requested_time_restriction)
            else:
                self.time_restriction = get_time_restriction(self.date_axis, self.time_axis,
                                                             min_year=min_year,
                                                             years=year,months=month,days=day,hours=hour,
                                                             previous=previous,next=next)
            #time sorting:
            self.time_restriction_sort = np.argsort(self.time_axis[self.time_restriction])
        else:
            self.time_axis,self.date_axis, self.time_restriction, self.time_restriction_sort=np.array([]),np.array([]),np.array([]),np.array([])

        #Set retrieveable variables:
        if 'soft_links' in self.data_root.groups:
            #Initialize variables:
            self.retrievable_vars = [var for var in self.data_root.variables 
                                     if  ( var in self.data_root.groups['soft_links'].variables and
                                           var != self.time_var)]

            #Get list of paths:
            for path_desc in ['path','path_id','file_type','version']+file_unique_id_list:
                setattr(self,path_desc+'_list',self.data_root.groups['soft_links'].variables[path_desc][:])
        else:
            self.retrievable_vars=[var for var in self.data_root.variables]

        #sort by size:
        size_retrievable_vars = [ _nonzeroprod(self.data_root.variables[var].shape) for var in self.retrievable_vars ]
        size_retrievable_vars_sort = np.argsort(size_retrievable_vars)[::-1]
        self.retrievable_vars = list(np.array(self.retrievable_vars)[size_retrievable_vars_sort])
        return

    def replicate(self,output,check_empty=False,chunksize=None):
        #replicate attributes
        netcdf_utils.replicate_netcdf_file(self.data_root,output)
        #replicate and copy variables:
        for var_name in self.data_root.variables:
            netcdf_utils.replicate_and_copy_variable(self.data_root, output, var_name, 
                                                     check_empty=check_empty, zlib=True, chunksize=chunksize)
        if 'soft_links' in self.data_root.groups:
            output_grp=netcdf_utils.replicate_group(self.data_root, output, 'soft_links')
            netcdf_utils.replicate_netcdf_file(self.data_root.groups['soft_links'], output_grp)
            for var_name in self.data_root.groups['soft_links'].variables:
                netcdf_utils.replicate_and_copy_variable(self.data_root.groups['soft_links'], output_grp, var_name,
                                                         check_empty=check_empty, zlib=True, chunksize=chunksize)
        return

    def append(self,output,check_empty=False):
        #replicate attributes
        netcdf_utils.replicate_netcdf_file(self.data_root, output)
    
        record_dimensions = netcdf_utils.append_record(self.data_root, output)
        #replicate and copy variables:
        for var_name in self.data_root.variables:
            if not var_name in record_dimensions:
                if ( var_name in output.variables and
                      netcdf_utils.check_dimensions_compatibility(self.data_root,
                                                                  output,
                                                                  var_name, exclude_unlimited=True) and
                      len(record_dimensions)>0 ):
                    #Variable can be appended along some record dimensions:
                    netcdf_utils.append_and_copy_variable(self.data_root, output, var_name, record_dimensions, check_empty=check_empty)
                elif ( not var_name in output.variables and 
                      netcdf_utils.check_dimensions_compatibility(self.data_root, output, var_name)):
                    #Variable can be copied:
                    netcdf_utils.replicate_and_copy_variable(self.data_root,output,var_name,check_empty=check_empty)
    
        if 'soft_links' in self.data_root.groups:
            data_grp = self.data_root.groups['soft_links']
            output_grp = netcdf_utils.replicate_group(self.data_root,output,'soft_links')
            netcdf_utils.replicate_netcdf_file(self.data_root.groups['soft_links'], output_grp)
    
            record_dimensions.update(netcdf_utils.append_record(data_grp, output_grp))
            for var_name in data_grp.variables:
                if not var_name in record_dimensions:
                    if ( var_name in output_grp.variables and
                          netcdf_utils.check_dimensions_compatibility(data_grp,output_grp,var_name,exclude_unlimited=True)):
                        #Variable can be appended along the time and path dimensions:
                        netcdf_utils.append_and_copy_variable(data_grp,output_grp,var_name,record_dimensions,check_empty=check_empty)
                    elif ( not var_name in output_grp.variables and 
                          netcdf_utils.check_dimensions_compatibility(data_grp,output_grp,var_name)):
                        #Variable can be copied:
                        netcdf_utils.replicate_and_copy_variable(data_grp,output_grp,var_name,check_empty=check_empty)
        return

    def retrieve(self,output,retrieval_type,filepath=None,out_dir='.'):
        #Define tree:
        self.tree=output.path.split('/')[1:]
        self.filepath=filepath
        self.out_dir=out_dir
        self.retrieval_type=retrieval_type

        if self.time_var!=None:
            #Record to output if output is a netCDF4 Dataset:
            if not self.time_var in output.dimensions:
                #pick only requested times and sort them
                netcdf_utils.create_time_axis(self.data_root,output,self.time_axis[self.time_restriction][self.time_restriction_sort],time_var=self.time_var)

            #Replicate all the other variables:
            for var in set(self.data_root.variables).difference(self.retrievable_vars):
                if not var in output.variables:
                    output = netcdf_utils.replicate_and_copy_variable(self.data_root,output,var)

            if self.retrieval_type in ['download_files','download_opendap']:
                #Replicate soft links for remote_queryable data:
                output_grp=netcdf_utils.replicate_group(self.data_root,output,'soft_links')
                for var_name in self.data_root.groups['soft_links'].variables:
                    netcdf_utils.replicate_netcdf_var(self.data_root.groups['soft_links'],output_grp,var_name)
                    if ( var_name != self.time_var and 
                         sum(self.time_restriction)>0 ):
                        if self.time_var in self.data_root.groups['soft_links'].variables[var_name].dimensions:
                            if self.data_root.groups['soft_links'].variables[var_name].shape[0]==1:
                                #Prevents a bug in h5py when self.data_root is an h5netcdf file:
                                if np.all(self.time_restriction):
                                    output_grp.variables[var_name][:] = (self.data_root.groups['soft_links']
                                                                                  .variables[var_name][...])
                            else:
                                output_grp.variables[var_name][:] = (self.data_root.groups['soft_links']
                                                                                  .variables[var_name]
                                                                                            [self.time_restriction,...]
                                                                                            [self.time_restriction_sort,...])
                        else:
                            output_grp.variables[var_name][:] = self.data_root.groups['soft_links'].variables[var_name][:]

            self.paths_sent_for_retrieval=[]
            for var_to_retrieve in self.retrievable_vars:
                self._retrieve_variable(output,var_to_retrieve)
        else:
            #Fixed variable. Do not retrieve, just copy:
            for var in self.retrievable_vars:
                output=netcdf_utils.replicate_and_copy_variable(self.data_root,output,var)
        return

    def _retrieve_variable(self,output,var_to_retrieve):
        #Replicate variable to output:
        output=netcdf_utils.replicate_netcdf_var(self.data_root,output,var_to_retrieve,chunksize=-1,zlib=True)

        if sum(self.time_restriction)==0:
            return

        #Get the requested dimensions:
        #self.get_dimensions_slicing()
        self.dimensions, self.unsort_dimensions=get_dimensions_slicing(self.data_root,var_to_retrieve,self.time_var)

        # Determine the paths_ids for soft links:
        if self.data_root.groups['soft_links'].variables[var_to_retrieve].shape[0]==1:
            #Prevents a bug in h5py when self.data_root is an h5netcdf file:
            if np.all(self.time_restriction):
                self.paths_link=self.data_root.groups['soft_links'].variables[var_to_retrieve][:,0]
                self.indices_link=self.data_root.groups['soft_links'].variables[var_to_retrieve][:,1]
        else:
            self.paths_link=self.data_root.groups['soft_links'].variables[var_to_retrieve][self.time_restriction,0][self.time_restriction_sort]
            self.indices_link=self.data_root.groups['soft_links'].variables[var_to_retrieve][self.time_restriction,1][self.time_restriction_sort]

        #Convert paths_link to id in path dimension:
        #self.paths_link=np.array([list(self.path_id_list).index(path_id) for path_id in self.paths_link])
        #Use search sorted:
        self.paths_link=np.argsort(self.path_id_list)[np.searchsorted(self.path_id_list,self.paths_link,
                                                                        sorter=np.argsort(self.path_id_list))]

        #Sort the paths so that we query each only once:
        unique_path_list_id, self.sorting_paths=np.unique(self.paths_link,return_inverse=True)

        for unique_path_id, path_id in enumerate(unique_path_list_id):
            self._retrieve_path_to_variable(unique_path_id,path_id,output,var_to_retrieve)
        return

    def _retrieve_path_to_variable(self,unique_path_id,path_id,output,var_to_retrieve):
        path_to_retrieve=self.path_list[path_id]

        #Next, we check if the file is available. If it is not we replace it
        #with another file with the same checksum, if there is one!
        file_type=self.file_type_list[list(self.path_list).index(path_to_retrieve)]
        if 'semaphores' in dir(self.q_manager):
            semaphores=self.q_manager.semaphores
        else:
            semaphores=dict()
        remote_data=remote_netcdf.remote_netCDF(path_to_retrieve,
                                                file_type,
                                                semaphores=semaphores,
                                                session=self.session,
                                                **self.remote_netcdf_kwargs)

        #See if the available path is available for download and find alternative:
        if self.retrieval_type=='download_files':
            path_to_retrieve=remote_data.check_if_available_and_find_alternative(self.path_list,self.file_type_list,self.checksum_list,remote_netcdf.downloadable_file_types,num_trials=2)
        elif self.retrieval_type=='download_opendap':
            path_to_retrieve=remote_data.check_if_available_and_find_alternative(self.path_list,self.file_type_list,self.checksum_list,remote_netcdf.remote_queryable_file_types,num_trials=2)
        elif self.retrieval_type=='load':
            path_to_retrieve=remote_data.check_if_available_and_find_alternative(self.path_list,self.file_type_list,self.checksum_list,remote_netcdf.local_queryable_file_types,num_trials=2)
        elif self.retrieval_type=='assign':
            path_to_retrieve=remote_data.check_if_available_and_find_alternative(self.path_list,self.file_type_list,self.checksum_list,remote_netcdf.queryable_file_types,num_trials=2)

        if path_to_retrieve is None:
            #Do not retrieve!
            return

        #See if there is a better file_type available:
        if self.retrieval_type=='download_files' and not self.download_all_files:
            alt_path_to_retrieve=remote_data.check_if_available_and_find_alternative(self.path_list,self.file_type_list,self.checksum_list,
                                                                remote_netcdf.remote_queryable_file_types+
                                                                remote_netcdf.local_queryable_file_types,num_trials=2)
            #Do not retrieve if a 'better' file type exists and is available
            if alt_path_to_retrieve!=None: return
        elif self.retrieval_type=='download_opendap' and not self.download_all_opendap:
            alt_path_to_retrieve=remote_data.check_if_available_and_find_alternative(self.path_list,self.file_type_list,self.checksum_list,remote_netcdf.local_queryable_file_types,num_trials=2)
            #Do not retrieve if a 'better' file type exists and is available
            if alt_path_to_retrieve!=None: return
        elif self.retrieval_type=='assign':
            path_to_retrieve=remote_data.check_if_available_and_find_alternative(self.path_list,self.file_type_list,self.checksum_list,remote_netcdf.local_queryable_file_types,num_trials=2)
            
        #Get the file_type, checksum and version of the file to retrieve:
        path_index=list(self.path_list).index(path_to_retrieve)
        file_type=self.file_type_list[path_index]
        version='v'+str(self.version_list[path_index])
        checksum=self.checksum_list[path_index]
        checksum_type=self.checksum_type_list[path_index]

        #Reverse pick time indices correponsing to the unique path_id:
        if file_type=='soft_links_container':
            #if the data is in the current file, the data lies in the corresponding time step:
            time_indices=np.arange(len(self.sorting_paths),dtype=int)[self.sorting_paths==unique_path_id]
        else:
            time_indices=self.indices_link[self.sorting_paths==unique_path_id]

        download_args=(0,path_to_retrieve,file_type,var_to_retrieve,self.tree)

        if self.retrieval_type!='download_files':
            #This is an important test that should be included in future releases:
            #with netCDF4.Dataset(path_to_retrieve.split('|')[0]) as data_test:
            #    data_date_axis=netcdf_utils.get_date_axis(data_test,'time')[time_indices]
            #print(path_to_retrieve,self.date_axis[self.time_restriction][self.time_restriction_sort][self.sorting_paths==unique_path_id],data_date_axis)
            self.dimensions[self.time_var], self.unsort_dimensions[self.time_var] = indices_utils.prepare_indices(time_indices)

            if self.retrieval_type=='download_opendap':
                new_path='soft_links_container/'+os.path.basename(self.path_list[path_index])
                new_file_type='soft_links_container'
                self._add_path_to_soft_links(new_path,new_file_type,path_index,self.sorting_paths==unique_path_id,output.groups['soft_links'],var_to_retrieve)

            sort_table=np.arange(len(self.sorting_paths))[self.sorting_paths==unique_path_id]
            download_kwargs={'dimensions':self.dimensions,
                             'unsort_dimensions':self.unsort_dimensions,
                             'sort_table':sort_table
                             }

            #Keep a list of paths sent for retrieval:
            self.paths_sent_for_retrieval.append(path_to_retrieve)
            if file_type=='soft_links_container':
                #The data is already in the file, we must copy it!
                max_request=2048
                retrieved_data=netcdf_utils.retrieve_container(self.data_root,
                                                                var_to_retrieve,
                                                                self.dimensions,
                                                                self.unsort_dimensions,
                                                                sort_table,max_request)
                result=(retrieved_data, sort_table, self.tree+[var_to_retrieve,])
                assign_leaf(output,*result)
            elif file_type in remote_netcdf.remote_queryable_file_types:
                data_node=remote_netcdf.get_data_node(path_to_retrieve,file_type)
                #Send to the download queue:
                self.q_manager.put_to_data_node(data_node,download_args+(download_kwargs,))
            elif file_type in remote_netcdf.local_queryable_file_types:
                #Load and simply assign:
                remote_data=remote_netcdf.remote_netCDF(path_to_retrieve,file_type)
                result=remote_data.download(*download_args[3:],download_kwargs=download_kwargs)
                assign_leaf(output,*result)
        else:
            if not path_to_retrieve in self.paths_sent_for_retrieval:
                new_path=http_netcdf.destination_download_files(self.path_list[path_index],
                                                                     self.out_dir,
                                                                     var_to_retrieve,
                                                                     self.path_list[path_index],
                                                                     self.tree)
                new_file_type='local_file'
                self._add_path_to_soft_links(new_path,new_file_type,path_index,self.sorting_paths==unique_path_id,output.groups['soft_links'],var_to_retrieve)
                download_kwargs={'out_dir':self.out_dir,
                                 'version':version,
                                 'checksum':checksum,
                                 'checksum_type':checksum_type}
                #Keep a list of paths sent for retrieval:
                self.paths_sent_for_retrieval.append(path_to_retrieve)

                data_node=remote_netcdf.get_data_node(path_to_retrieve,file_type)
                #Send to the download queue:
                self.q_manager.put_to_data_node(data_node,download_args+(download_kwargs,))
        return 

    def _add_path_to_soft_links(self,new_path,new_file_type,path_index,time_indices_to_replace,output,var_to_retrieve):
        if not new_path in output.variables['path'][:]:
            output.variables['path'][len(output.dimensions['path'])]=new_path
            output.variables['path_id'][-1]=hash(new_path)
            output.variables['file_type'][-1]=new_file_type
            output.variables['data_node'][-1]=remote_netcdf.get_data_node(new_path,output.variables['file_type'][-1])
            for path_desc in ['version']+file_unique_id_list:
                output.variables[path_desc][-1]=getattr(self,path_desc+'_list')[path_index]
        
        output.variables[var_to_retrieve][time_indices_to_replace,0]=output.variables['path_id'][-1]
        return output

    def open(self):
        self.tree=[]
        filehandle,self.filepath=tempfile.mkstemp()
        self.output_root=netCDF4.Dataset(self.filepath,
                                      'w',format='NETCDF4',diskless=True,persist=False)
        self._is_open=True
        return

    def __enter__(self):
        self.open()
        return self

    def assign(self,var_to_retrieve,q_manager=None):
        if not self._is_open:
            raise IOError('read_soft_lnks must be opened to assign')

        self.variables=dict()
        #Create type download_opendap_and_load!
        self.retrieval_type='assign'
        self.out_dir='.'
        self.paths_sent_for_retrieval=[]
     
        self.output_root.createGroup(var_to_retrieve)
        netcdf_utils.create_time_axis(self.data_root,self.output_root.groups[var_to_retrieve],self.time_axis[self.time_restriction][self.time_restriction_sort])
        self._retrieve_variable(self.output_root.groups[var_to_retrieve],var_to_retrieve)
       
        for var in self.output_root.groups[var_to_retrieve].variables:
            self.variables[var]=self.output_root.groups[var_to_retrieve].variables[var]
        return

    def close(self):
        self.output_root.close()
        try:
            os.remove(self.filepath)
        except:
            pass
        self._is_open=False
        return

    def __exit__(self, *_):
        self.close()
        return

def add_previous(time_restriction):
    return np.logical_or(time_restriction,np.append(time_restriction[1:],False))

def add_next(time_restriction):
    return np.logical_or(time_restriction,np.insert(time_restriction[:-1],0,False))

def time_restriction_years(min_year,years,date_axis,time_restriction_any):
    if years!=None:
        years_axis=np.array([date.year for date in date_axis])
        if min_year!=None:
            #Important for piControl:
            time_restriction=np.logical_and(time_restriction_any, [True if year in years else False for year in years_axis-years_axis.min()+min_year])
        else:
            time_restriction=np.logical_and(time_restriction_any, [True if year in years else False for year in years_axis])
        return time_restriction
    else:
        return time_restriction_any

def time_restriction_months(months,date_axis,time_restriction_for_years):
    if months!=None:
        months_axis=np.array([date.month for date in date_axis])
        #time_restriction=np.logical_and(time_restriction,months_axis==month)
        time_restriction=np.logical_and(time_restriction_for_years,[True if month in months else False for month in months_axis])
        return time_restriction
    else:
        return time_restriction_for_years

def time_restriction_days(days,date_axis,time_restriction_any):
    if days!=None:
        days_axis=np.array([date.day for date in date_axis])
        time_restriction=np.logical_and(time_restriction_any,[True if day in days else False for day in days_axis])
        return time_restriction
    else:
        return time_restriction_any
                    
def time_restriction_hours(hours,date_axis,time_restriction_any):
    if hours!=None:
        hours_axis=np.array([date.hour for date in date_axis])
        time_restriction=np.logical_and(time_restriction_any,[True if hour in hours else False for hour in hours_axis])
        return time_restriction
    else:
        return time_restriction_any
                    
def get_time_restriction(date_axis, time_axis,
                         min_year=None, years=None, months=None, days=None, hours=None,
                         previous=0,next=0):
    time_restriction = np.ones(date_axis.shape,dtype=np.bool)

    time_restriction = time_restriction_years(min_year,years,date_axis,time_restriction)
    time_restriction = time_restriction_months(months,date_axis,time_restriction)
    time_restriction = time_restriction_days(days,date_axis,time_restriction)
    time_restriction = time_restriction_hours(hours,date_axis,time_restriction)

    if ( (previous>0) or
         (next>0) ):
        sort_indices = np.argsort(time_axis)

        sorted_time_restriction = time_restriction[sort_indices]
        if previous>0:
            for prev_num in range(previous):
                sorted_time_restriction = add_previous(sorted_time_restriction)
        if next>0:
            for next_num in range(next):
                sorted_time_restriction = add_next(sorted_time_restriction)
        time_restriction[sort_indices] = sorted_time_restriction
    return time_restriction

def get_dimensions_slicing(dataset,var,time_var):
    #Set the dimensions:
    dimensions=dict()
    unsort_dimensions=dict()
    for dim in dataset.variables[var].dimensions:
        if dim != time_var:
            if dim in dataset.variables:
                dimensions[dim] = dataset.variables[dim][:]
            else:
                dimensions[dim] = np.arange(len(dataset.dimensions[dim]))
            unsort_dimensions[dim] = None
    return dimensions, unsort_dimensions

def assign_leaf(output, val, sort_table, tree):
    output.variables[tree[-1]][sort_table, ...] = np.ma.masked_invalid(val)
    return



#External:
import numpy as np
from itertools import groupby, count

def get_indices_from_dim(source,output):
    #This function finds which indices from source should be used and in which order:

    #Special case where both dimensions are lenght one. Then it is unambiguous:
    if len(source)==1 and len(output)==1:
        return np.array([0,])

    indices=np.arange(max(source.shape))[np.in1d(source,output)]
    try:
        return np.array([ indices[source[indices]==val][0] for val in output ])
    except IndexError:
        #The in1d might have encountered afloating point error. Make the equality fuzzy:
        #warnings.warn('Dimension matching was done to floating point tolerance for some model',UserWarning)
        indices=np.arange(max(source.shape))[np.array([np.any(np.isclose(output,item,atol=1e-5)) for item in source])]
        #try:
        return np.array([ indices[np.isclose(source[indices],val,atol=1e-5)][0] for val in output ])
        #except IndexError:
        #    print indices
        #    print(source,output)
        #    raise

def convert_indices_to_slices(indices):
    #This feature is currently broken (December 2014):
    if len(indices)>1:
        steps=np.maximum(np.unique(np.diff(indices)),1)
        optimal_step=steps[np.argmin(map(lambda x: len(convert_indices_to_slices_step(indices,x)),steps))]
        return convert_indices_to_slices_step(indices,optimal_step)
    else:
        return convert_indices_to_slices_step(indices,1)

def convert_indices_to_slices_step(indices,step):
    slices = []
    for key, it in groupby(enumerate(indices), lambda x: x[1] - step*x[0]):
        indices_slice = [y for x, y in it]
        slices.append(slice(indices_slice[0], indices_slice[-1]+1,step))
        #if len(indices) == 1:
        #    slices.append(slice(indices[0],indices[0]+1))
        #else:
        #    slices.append(slice(indices[0], indices[-1]+1))
    return slices

def slice_a_slice(initial_slice,slice_to_use):
    """
    This function is not commutative
    """
    if isinstance(slice_to_use,slice) and isinstance(initial_slice,slice):
        return convert_indices_to_slices_step(range(initial_slice.start,initial_slice.stop,initial_slice.step)[slice_to_use],initial_slice.step*slice_to_use.step)[0]
    elif isinstance(initial_slice,slice):
        return range(initial_slice.start,initial_slice.stop,initial_slice.step)[slice_to_use]
    else:
        return np.array(initial_slice)[slice_to_use]
        

def prepare_indices(indices):
    sort_indices=np.argsort(indices)
    #Sort:
    indices=indices[sort_indices]
    #provide the inverse:
    unsort_indices=np.argsort(sort_indices)

    #always retrieve the first index (bug in netCDF4 python):
    #if not 0 in indices:
    #    indices=np.insert(indices[sort_indices],0,0,axis=0)
    #    unsort_indices+=1

    #Finally, convert the indices to slices:
    indices=convert_indices_to_slices(indices)
    return indices, unsort_indices

def largest_hyperslab(slices_dict):
    return np.prod([max([slice_length(item) for item in slices_dict[dim]])
                for dim in slices_dict.keys()])

def slice_length(slice_item):
    return len(range(slice_item.start,slice_item.stop,slice_item.step))
        
def retrieve_slice(variable,indices,unsort_indices,dim,dimensions,dim_id,max_request,getitem_tuple=tuple(),default=False):
    if default: return np.array([])
    if len(dimensions)>0:
        return take_safely(np.ma.concatenate(map(lambda x: retrieve_slice(variable,
                                                 indices,
                                                 unsort_indices,
                                                 dimensions[0],
                                                 dimensions[1:],
                                                 dim_id+1,
                                                 max_request,
                                                 getitem_tuple=getitem_tuple+(x,)),
                                                 indices[dim]),
                              axis=dim_id),unsort_indices[dim],axis=dim_id)
    else:
        return take_safely(np.ma.concatenate(map(lambda x: getitem_from_variable(variable,getitem_tuple+(x,),max_request),
                                                 indices[dim]),
                              axis=dim_id),unsort_indices[dim],axis=dim_id)

def take_safely(x, indices, axis=0):
    if x.shape[axis] == 0:
        axes = len(x.shape) * [None]
        for id in range(len(axes)):
            axes[id] = x.shape[id]
        axes[axis] = len(indices)
        return np.ma.masked_all(axes)
    else:
        return np.ma.take(x, indices, axis=axis)

def getitem_from_variable(variable,getitem_tuple,max_request):
    if ( max_request==None or
         max_request*1024*1024>32*np.prod([( (item.stop-item.start)//item.step) for item in getitem_tuple]) ):
        
        return variable[getitem_tuple]
    else:
        #Max number of steps along first dimension: 
        max_steps = np.maximum(int(np.floor(max_request*1024*1024/(32*np.prod([( (item.stop-item.start)//item.step) for item in getitem_tuple[1:]])))), 1)

        first_dim_length = (getitem_tuple[0].stop - getitem_tuple[0].start) // getitem_tuple[0].step
        num_split = np.minimum( first_dim_length // max_steps, first_dim_length)

        id_lists = np.array_split(np.arange( getitem_tuple[0].start,  getitem_tuple[0].stop,  getitem_tuple[0].step), num_split)

        slice_list = map(lambda x: convert_indices_to_slices_step(x, getitem_tuple[0].step)[0], id_lists)
        return np.ma.concatenate(map(lambda x: variable[(x,) + getitem_tuple[1:]], slice_list),
                                 axis=0)
                                            



#External:
import multiprocessing
import numpy as np
import Queue
import requests
import requests_cache

#Internal:

#http://eli.thegreenplace.net/2012/01/04/shared-counter-with-pythons-multiprocessing
class Shared_Counter(object):
    #Shared counter class
    def __init__(self,manager):
        self.val = manager.Value('i', 0)
        self.lock = manager.Lock()

    def increment(self,n=1):
        with self.lock:
            return self.increment_no_lock(n=n)

    def increment_no_lock(self,n=1):
        value=self.val.value
        self.val.value += n
        return value + n

    def decrement(self,n=1):
        with self.lock:
            return self.decrement_no_lock(n=n)

    def decrement_no_lock(self,n=1):
        value=self.val.value
        self.val.value -= n
        return value - n

    @property
    def value(self):
        with self.lock:
            return self.val.value

    @property
    def value_no_lock(self):
        return self.val.value

class Semaphores_data_node:
    #Shared semaphores class
    def __init__(self,manager,n=20,num_concurrent=1):
        self.dict = manager.dict()
        self.lock = manager.Lock()
        self.list = [manager.Semaphore(num_concurrent) for id in range(n)]

    def add_new_data_node(self,data_node): 
        #Add a pointer to a semaphore:
        with self.lock:
            if not data_node in self.dict.keys():
                if len(self.dict.keys())>0:
                    self.dict[data_node]=1+np.max([self.dict[key] for key in self.dict.keys()])
                else:
                    self.dict[data_node]=0
        return

    def __getitem__(self,data_node):
        #Just send the semaphore object:
        return self.list[self.dict[data_node]]

    def keys(self):
        #Just send the semaphore object:
        return self.dict.keys()

class Queues_data_node:
    #Shared queues class
    def __init__(self,manager,n=20):
        self.dict = manager.dict()
        self.lock = manager.Lock()
        self.list = [manager.Queue() for id in range(n)]
        self.list_expected = [Shared_Counter(manager) for id in range(n)]

    def add_new_data_node(self,data_node): 
        #Add a pointer to a queue:
        with self.lock:
            if not data_node in self.dict.keys():
                if len(self.dict.keys())>0:
                    self.dict[data_node]=1+np.max([self.dict[key] for key in self.dict.keys()])
                else:
                    self.dict[data_node]=0
        return

    def put(self,data_node,item):
        #Just put the item in the queue and increment the counter:
        with self.list_expected[self.dict[data_node]].lock:
            self.list[self.dict[data_node]].put(item)
            self.list_expected[self.dict[data_node]].increment_no_lock()
        return 

    def get(self,data_node):
        #Just send the queue object:
        item=self.list[self.dict[data_node]].get()
        #Decrement after, in case there is an error in the get command:
        self.list_expected[self.dict[data_node]].decrement()
        return item

    def qsize(self,data_node):
        return self.list_expected[self.dict[data_node]].value

    def keys(self):
        return self.dict.keys()

class NC4SL_queues_manager:
    def __init__(self,options,processes_names,manager=None,remote_netcdf_kwargs=dict()):
        if manager==None:
            self.manager=multiprocessing.Manager()
        else:
            self.manager=manager

        #Shared sessions among downloads. Appears tricky to use...
        #self.session=requests_sessions.create_single_session(**remote_netcdf_kwargs)
        ##Set the pool size to number of downloads:
        #adapter = requests.adapters.HTTPAdapter(pool_connections=20, pool_maxsize=options.num_dl)
        #self.session.mount('http://',adapter)
        #self.session.mount('https://',adapter)

        self.semaphores=Semaphores_data_node(self.manager,num_concurrent=options.num_dl)
        self.queues=Queues_data_node(self.manager)
        #Create gather download_queues:
        for proc_id in processes_names:
            thread_id='download_'+proc_id
            setattr(self,thread_id, self.manager.Queue())
            setattr(self,thread_id+'_expected', Shared_Counter(self.manager))
            setattr(self,thread_id+'_closed', self.manager.Event())
        return
                
    def put_to_data_node(self,data_node,item):
        thread_id='download_'+multiprocessing.current_process().name
        #thread_id='download_'+process_name
        self.put_to_data_node_from_thread_id(thread_id,data_node,item)
        return

    def put_to_data_node_from_thread_id(self,thread_id,data_node,item):
        #Increment expected counter
        with getattr(self,thread_id+'_expected').lock:
            #Pass the thread_id to data_node queue:
            self.queues.put(data_node,(thread_id,)+item)
            getattr(self,thread_id+'_expected').increment_no_lock()
        return

    def put_again_to_data_node_from_thread_id(self,thread_id,data_node,item):
        #The expectation has already been recorded:
        self.queues.put(data_node,(thread_id,)+item)
        return

    def set_closed(self):
        thread_id='download_'+multiprocessing.current_process().name
        #thread_id='download_'+process_name
        getattr(self,thread_id+'_closed').set()
        return

    def set_opened(self):
        thread_id='download_'+multiprocessing.current_process().name
        #thread_id='download_'+process_name
        getattr(self,thread_id+'_closed').clear()
        return

    def put_for_thread_id(self,thread_id,item):
        getattr(self,thread_id).put(item)
        return

    def get_for_thread_id(self):
        timeout=0.1
        thread_id='download_'+multiprocessing.current_process().name
        #thread_id='download_'+process_name
        while not (getattr(self,thread_id+'_closed').is_set() 
                   and getattr(self,thread_id+'_expected').value==0):
            try:
                with getattr(self,thread_id+'_expected').lock:
                    item = getattr(self,thread_id).get(True,timeout)
                    #Decrement expected counter
                    getattr(self,thread_id+'_expected').decrement_no_lock()
                return item
            except Queue.Empty:
                pass
        return 'STOP'



#External:
import sys

#Internal:
from .certificates import certificates
from . import commands, parsers

def main(string_call=None):
    import argparse 
    import textwrap

    #Generate subparsers
    parser = parsers.generate_subparsers(None)

    if string_call != None:
        options = parser.parse_args(string_call)
    else:
        options = parser.parse_args()

    options = certificates.prompt_for_username_and_password(options)

    if options.command != 'certificates':
        getattr(commands, options.command)(options)
        
if __name__ == "__main__":
    main()



#External:
import time
import os
import errno
import datetime
import hashlib

from socket import error as SocketError
import requests
from netcdf4_pydap.httpserver import Dataset, RemoteEmptyError

#Internal:
from . import safe_handling

class http_netCDF:
    def __init__(self,url,semaphores=dict(),
                               remote_data_node='',
                               timeout=120,
                               cache=None,
                               expire_after=datetime.timedelta(hours=1),
                               session=None,
                               authentication_url=None,
                               username=None,
                               password=None,
                               use_certificates=False):
        self.url=url
        self.semaphores=semaphores
        self.timeout=timeout
        self.cache=cache
        self.expire_after=expire_after
        self.session=session
        self.authentication_url=authentication_url
        self.username=username
        self.password=password
        self.use_certificates=use_certificates

        if (remote_data_node in  self.semaphores.keys()):
            self.semaphore=semaphores[remote_data_node]
            self.handle_safely=True
        else:
            self.semaphore=safe_handling.dummy_semaphore()
            self.handle_safely=False
        return

    def __enter__(self):
        self.semaphore.acquire()
        return self

    def __exit__(self,type,value,traceback):
        if self.handle_safely:
            #Do not release semaphore right away if data is not local:
            time.sleep(0.01)
        self.semaphore.release()
        return

    def check_if_opens(self,num_trials=5):
        #If ftp, assume available:
        if len(self.url)>3 and self.url[:3]=='ftp':
            return True
        success=False
        for trial in range(num_trials):
            if not success:
                try:
                    with Dataset(self.url,
                                 cache=self.cache,
                                 timeout=self.timeout,
                                 expire_after=self.expire_after,
                                 session=self.session,
                                 authentication_url=self.authentication_url,
                                 username=self.username,
                                 password=self.password,
                                 use_certificates=self.use_certificates) as dataset:
                        pass
                    success=True
                except requests.exceptions.ReadTimeout as e:
                    time.sleep(3*(trial+1))
                    pass
                except HTTPError as e:
                    time.sleep(3*(trial+1))
                    pass
                except URLError as e:
                    if e.message == '<urlopen error [Errno 110] Connection timed out>':
                        time.sleep(3*(trial+1))
                        pass
                    else:
                        raise
                except requests.exceptions.ConnectionError as e:
                    time.sleep(3*(trial+1))
                    pass
                except SocketError as e:
                    #http://stackoverflow.com/questions/20568216/python-handling-socket-error-errno-104-connection-reset-by-peer
                    if e.errno != errno.ECONNRESET:
                        raise
                    time.sleep(3*(trial+1))
                    pass
                except RemoteEmptyError as e:
                    print(e)
                    break
        return success

    def download(self,var,pointer_var,checksum='',checksum_type='MD5',out_dir='.',version='v1'):
        dest_name=destination_download_files(self.url,out_dir,var,version,pointer_var)

        if checksum=='':
            if os.path.isfile(dest_name):
                return 'File '+dest_name+' found but could NOT check checksum of existing file because checksum was not a priori available. Not retrieving.'
        else:
            try: #Works only if file exists!
                comp_checksum=checksum_for_file(checksum_type,dest_name)
            except:
                comp_checksum=''
            if comp_checksum==checksum:
                return 'File '+dest_name+' found. '+checksum_type+' OK! Not retrieving.'

        with Dataset(self.url,
                     cache=self.cache,
                     timeout=self.timeout,
                     expire_after=self.expire_after,
                     session=self.session,
                     authentication_url=self.authentication_url,
                     username=self.username,
                     password=self.password,
                     use_certificates=self.use_certificates) as dataset:
           
            size_string=dataset.wget(dest_name,progress=True)

        if checksum=='':
            return size_string+'\n'+'Could NOT check checksum of retrieved file because checksum was not a priori available.'
        else:
            try: 
                comp_checksum=checksum_for_file(checksum_type,dest_name)
            except:
                comp_checksum=''
            if comp_checksum!=checksum:
                try:
                    os.remove(dest_name)
                except:
                    pass
                return size_string+'\n'+'File '+dest_name+' does not have the same '+checksum_type+' checksum as published on the ESGF. Removing this file...'
            else:
                return size_string+'\n'+'Checking '+checksum_type+' checksum of retrieved file... '+checksum_type+' OK!'

def destination_download_files(url,out_dir,var,version,pointer_var):
    if out_dir.split('/')[-2]=='tree':
        dest_name=out_dir.replace('tree','/'.join(pointer_var))+'/'
    else:
        dest_name=out_dir.replace('tree','/'.join(pointer_var[:-1]))+'/'
        dest_name=dest_name.replace('var',var)
        dest_name=dest_name.replace('version',version)

    dest_name+=url.split('/')[-1]
    return os.path.abspath(os.path.expanduser(os.path.expandvars(dest_name)))

def md5_for_file(f, block_size=2**20):
    md5 = hashlib.md5()
    while True:
        data = f.read(block_size)
        if not data:
            break
        md5.update(data)
    return md5.hexdigest()

def sha_for_file(f, block_size=2**20):
    sha = hashlib.sha256()
    while True:
        data = f.read(block_size)
        if not data:
            break
        sha.update(data)
    return sha.hexdigest()

def checksum_for_file(checksum_type,dest_name, block_size=2**20):
    checksum = getattr(hashlib,checksum_type.lower())()
    with open(dest_name,'rb') as f:
        while True:
            data = f.read(block_size)
            if not data:
                break
            checksum.update(data)
    return checksum.hexdigest()



#External
import datetime
import os
import netCDF4
import socket
import multiprocessing
import requests

#Internal
from .soft_links import create_soft_links, read_soft_links
from .subset import subset_utils
from .remote_netcdf import remote_netcdf
from . import retrieval_manager, queues_manager

valid_file_type_list=['local_file','OPENDAP']
time_frequency=None
unique_file_id_list=['checksum_type','checksum','tracking_id']
drs_to_pass=['path','version','file_type','data_node']

def validate(options):
    input_paths=options.in_netcdf_file
    version=datetime.datetime.now().strftime('%Y%m%d')

    if options.file_type=='local_file':
        hostname = socket.gethostname()
        input_paths=[ ':'.join([hostname,os.path.abspath(os.path.expanduser(os.path.expandvars(path)))]) for path in input_paths ]
    data_node_list=[remote_netcdf.get_data_node(path,options.file_type) for path in input_paths]

    paths = zip([ '|'.join([path,]+['' for id in unique_file_id_list]) for path in input_paths],
                [version for path in input_paths],
                [options.file_type for path in input_paths],
                data_node_list)
    paths_list=[{drs_name:path[drs_id] for drs_id, drs_name in enumerate(drs_to_pass)} for path in paths]

    manager=multiprocessing.Manager()
    validate_semaphores=queues_manager.Semaphores_data_node(manager,num_concurrent=5)

    remote_netcdf_kwargs={opt: getattr(options,opt) for opt in ['openid','username','password','use_certificates',
                                                                 ] if opt in dir(options)}
    session=requests.Session()
    netcdf_pointers=create_soft_links.create_netCDF_pointers(
                                                      paths_list,
                                                      time_frequency,options.year,options.month,
                                                      valid_file_type_list,
                                                      list(set(data_node_list)),
                                                      record_other_vars=False,
                                                      semaphores=validate_semaphores,
                                                      time_var=options.time_var,
                                                      session=session,
                                                      remote_netcdf_kwargs=remote_netcdf_kwargs)
    output=netCDF4.Dataset(options.out_netcdf_file,'w')
    netcdf_pointers.record_meta_data(output,options.var_name)
    return

def download_files(options):
    download(options,retrieval_type='download_files')
    return

def download_opendap(options):
    download(options,retrieval_type='download_opendap')
    return

def load(options):
    download(options,retrieval_type='load')
    return

def download(options,retrieval_type='load'):

    output=netCDF4.Dataset(options.out_netcdf_file,'w')
    data=netCDF4.Dataset(options.in_netcdf_file,'r')
    data_node_list=list(set(data.groups['soft_links'].variables['data_node'][:]))

    if retrieval_type!='load':
        #Create manager:
        processes_names=[multiprocessing.current_process().name,]
        q_manager=queues_manager.NC4SL_queues_manager(options,processes_names)

        #Create download queues:
        for data_node in data_node_list:
            q_manager.semaphores.add_new_data_node(data_node)
            q_manager.queues.add_new_data_node(data_node)

        download_processes=retrieval_manager.start_download_processes(options,q_manager)

    try:
        q_manager.set_opened()
        remote_netcdf_kwargs={opt: getattr(options,opt) for opt in ['openid','username','password','use_certificates',
                                                                     ] if opt in dir(options)}
        session=requests.Session()
        options_dict={opt: getattr(options,opt) for opt in ['previous','next','year','month','day','hour',
                                                                     'download_all_files','download_all_opendap'] if opt in dir(options)}
        options_dict['remote_netcdf_kwargs']=remote_netcdf_kwargs

        netcdf_pointers=read_soft_links.read_netCDF_pointers(data,time_var=options.time_var,
                                                                  q_manager=q_manager,
                                                                  session=session,
                                                                  **options_dict)
        if retrieval_type=='download_files':
            netcdf_pointers.retrieve(output,retrieval_type,filepath=options.out_netcdf_file,out_dir=options.out_download_dir)
        else:
            netcdf_pointers.retrieve(output,retrieval_type,filepath=options.out_netcdf_file)

        if retrieval_type!='load':
            #Close queues:
            q_manager.set_closed()
            output=retrieval_manager.launch_download(output,data_node_list,q_manager,options)
            output.close()
            if ( retrieval_type=='download_files' and
                not ( 'do_not_revalidate' in dir(options) and options.do_not_revalidate)):
                pass
                #Revalidate not implemented yet
    finally:
        if retrieval_type!='load':
            #Terminate the download processes:
            for item in download_processes.keys():
                download_processes[item].terminate()
    return

def subset(options):
    subset_utils.subset(options.in_netcdf_file,options.out_netcdf_file,
                            lonlatbox=options.lonlatbox,
                            lat_var=options.lat_var,lon_var=options.lon_var,
                            output_vertices=options.output_vertices)
    return



#!/usr/bin/env python
"""
   :platform: Unix
   :synopsis: Rewrite and/or check time axis of MIP NetCDF files. Modified by F. Laliberte to work with cdb_query

"""

# Module imports
import re
import os
import argparse
import logging
import ConfigParser
from uuid import uuid4
import numpy as np
from argparse import RawTextHelpFormatter
from datetime import datetime
from functools import wraps
from netCDF4 import Dataset, date2num, num2date
from netCDF4 import datetime as phony_datetime
from multiprocessing.dummy import Pool as ThreadPool
from textwrap import fill
from glob import glob

# Program version
__version__ = '{0} {1}-{2}-{3}'.format('v3.0', '2015', '08', '25')


class ProcessingContext(object):
    """
    Encapsulates the following processing context/information for main process:

    +-------------------+-------------+---------------------------+
    | Attribute         | Type        | Description               |
    +===================+=============+===========================+
    | *self*.directory  | *str*       | Variable to scan          |
    +-------------------+-------------+---------------------------+
    | *self*.check      | *boolean*   | True if check mode        |
    +-------------------+-------------+---------------------------+
    | *self*.write      | *boolean*   | True if write mode        |
    +-------------------+-------------+---------------------------+
    | *self*.force      | *boolean*   | True if force writing     |
    +-------------------+-------------+---------------------------+
    | *self*.project    | *str*       | Project                   |
    +-------------------+-------------+---------------------------+
    | *self*.checksum   | *str*       | The checksum type         |
    +-------------------+-------------+---------------------------+
    | *self*.cfg        | *callable*  | Configuration file parser |
    +-------------------+-------------+---------------------------+
    | *self*.attributes | *dict*      | DRS attributes from regex |
    +-------------------+-------------+---------------------------+
    | *self*.pattern    | *re object* | Filename regex pattern    |
    +-------------------+-------------+---------------------------+
    | *self*.calendar   | *str*       | NetCDF calendar attribute |
    +-------------------+-------------+---------------------------+
    | *self*.tunits     | *str*       | Time units from file      |
    +-------------------+-------------+---------------------------+
    | *self*.funits     | *str*       | Time units from frequency |
    +-------------------+-------------+---------------------------+

    :param dict args: Parsed command-line arguments
    :returns: The processing context
    :rtype: *dict*
    :raises Error: If the project name is inconsistent with the sections names from the configuration file
    :raises Error: If the regular expression from the configurtion file cannot match the supplied directory

    """
    def __init__(self, args):
        init_logging(args.logdir)
        self.directory = check_directory(args.directory)
        self.check = args.check
        self.write = args.write
        self.force = args.force
        self.verbose = args.verbose
        self.cfg = config_parse(args.config)
        if args.project in self.cfg.sections():
            self.project = args.project
        else:
            raise Exception('No section in configuration file corresponds to "{0}" project. Supported projects are {1}.'.format(args.project, self.cfg.sections()))
        self.pattern = re.compile(self.cfg.get(self.project, 'filename_format'))
        try:
            self.attributes = re.match(re.compile(self.cfg.get(self.project, 'directory_format')), self.directory).groupdict()
        except:
            # Fails can be due to:
            # -> Wrong project argument
            # -> Mistake in directory_format pattern in configuration file
            # -> Wrong version format (defined as one or more digit after 'v')
            # -> Wrong ensemble format (defined as r[0-9]i[0-9]p[0-9])
            # -> Mistake in the DRS tree of your filesystem.
            raise Exception('The "directory_format" regex cannot match {0}'.format(self.directory))
        self.checksum = str(self.cfg.defaults()['checksum_type'])
        self.calendar = None
        self.tunits = None
        self.funits = None


class AxisStatus(object):
    """
    Encapsulates the following file diagnostic to print or save:

    +------------------+-----------+---------------------------------------------------+
    | Attribute        | Type      | Description                                       |
    +==================+===========+===================================================+
    | *self*.directory | *str*     | Variable to scan                                  |
    +------------------+-----------+---------------------------------------------------+
    | *self*.file      | *str*     | Filename                                          |
    +------------------+-----------+---------------------------------------------------+
    | *self*.start     | *str*     | Start date from filename                          |
    +------------------+-----------+---------------------------------------------------+
    | *self*.end       | *str*     | End date from filename                            |
    +------------------+-----------+---------------------------------------------------+
    | *self*.last      | *str*     | Last date from theoretical time axis              |
    +------------------+-----------+---------------------------------------------------+
    | *self*.steps     | *int*     | Time axis length                                  |
    +------------------+-----------+---------------------------------------------------+
    | *self*.instant   | *boolean* | True if instantaneous time axis                   |
    +------------------+-----------+---------------------------------------------------+
    | *self*.calendar  | *str*     | NetCDF calendar attribute                         |
    +------------------+-----------+---------------------------------------------------+
    | *self*.units     | *str*     | Expected time units                               |
    +------------------+-----------+---------------------------------------------------+
    | *self*.control   | *list*    | Errors status                                     |
    +------------------+-----------+---------------------------------------------------+
    | *self*.bnds      | *boolean* | True if time boundaries excpeted and not mistaken |
    +------------------+-----------+---------------------------------------------------+
    | *self*.checksum  | *str*     | New checksum if modified file                     |
    +------------------+-----------+---------------------------------------------------+
    | *self*.axis      | *array*   | Theoretical time axis                             |
    +------------------+-----------+---------------------------------------------------+
    | *self*.time      | *array*   | Time axis from file                               |
    +------------------+-----------+---------------------------------------------------+

    :returns: The axis status
    :rtype: *dict*

    """
    def __init__(self):
        self.directory = None
        self.file = None
        self.start = None
        self.end = None
        self.last = None
        self.steps = None
        self.instant = False
        self.calendar = None
        self.units = None
        self.control = []
        self.bnds = None
        self.checksum = None
        self.axis = None
        self.time = None


def get_args(job):
    """
    Returns parsed command-line arguments. See ``time_axis -h`` for full description.
    A ``job`` dictionnary can be used as developper's entry point to overload the parser.

    :param dict job: Optionnal dictionnary instead of command-line arguments.
    :returns: The corresponding ``argparse`` Namespace

    """
    parser = argparse.ArgumentParser(
        description="""Rewrite and/or check time axis into MIP NetCDF files, considering\n(i) uncorrupted filename period dates and\n(ii) properly-defined times units, time calendar and frequency NetCDF attributes.\n\nReturned status:\n000: Unmodified time axis,\n001: Corrected time axis because wrong timesteps.\n002: Corrected time axis because of changing time units,\n003: Ignored time axis because of inconsistency between last date of time axis and\nend date of filename period (e.g., wrong time axis length),\n004: Corrected time axis deleting time boundaries for instant time,\n005: Ignored averaged time axis without time boundaries.""",
        formatter_class=RawTextHelpFormatter,
        add_help=False,
        epilog="""Developped by Levavasseur, G. (CNRS/IPSL) and Laliberte, F. (ExArch)""")
    parser.add_argument(
        'directory',
        nargs='?',
        help="""Variable path to diagnose following the DRS\n(e.g., /path/to/your/archive/CMIP5/merge/NCAR/CCSM4/amip/day/atmos/cfDay/r7i1p1/v20130507/tas/).\n\n"""),
    parser.add_argument(
        '-p', '--project',
        type=str,
        required=True,
        help="""Required project name corresponding to a section of the configuration file.\n\n""")
    parser.add_argument(
        '-C', '--config',
        type=str,
        default='{0}/config.ini'.format(os.path.dirname(os.path.abspath(__file__))),
        help="""Path of configuration INI file\n(default is {0}/config.ini).\n\n""".format(os.path.dirname(os.path.abspath(__file__))))
    parser.add_argument(
        '-h', '--help',
        action="help",
        help="""Show this help message and exit.\n\n""")
    group = parser.add_mutually_exclusive_group(required=True)
    group.add_argument(
        '-c', '--check',
        action='store_true',
        default=True,
        help='Check time axis squareness (default is True).\n\n')
    group.add_argument(
        '-w', '--write',
        action='store_true',
        default=False,
        help="""Rewrite time axis depending on checking\n(includes --check ; default is False).\nTHIS ACTION DEFINITELY MODIFY INPUT FILES!\n\n""")
    group.add_argument(
        '-f', '--force',
        action='store_true',
        default=False,
        help="""Force time axis writing overpassing checking step\n(default is False).\nTHIS ACTION DEFINITELY MODIFY INPUT FILES!\n\n""")
    parser.add_argument(
        '-l', '--logdir',
        type=str,
        nargs='?',
        const=os.getcwd(),
        help="""Logfile directory (default is working directory).\nIf not, standard output is used.\n\n""")
    parser.add_argument(
        '-v', '--verbose',
        action='store_true',
        default=False,
        help='Verbose mode.\n\n')
    parser.add_argument(
        '-V', '--version',
        action='version',
        version='%(prog)s ({0})'.format(__version__),
        help="""Program version.""")
    if job is None:
        return parser.parse_args()
    else:
        return parser.parse_args([job['full_path_variable'], '-p', job['project'], '-C', '/prodigfs/esg/ArchiveTools/sdp/conf/time_axis.ini', '-l', 'synda_logger', '-c', '-v'])


def init_logging(logdir):
    """
    Initiates the logging configuration (output, message formatting). In the case of a logfile, the logfile name is unique and formatted as follows:
    name-YYYYMMDD-HHMMSS-PID.log

    :param str logdir: The relative or absolute logfile directory. If ``None`` the standard output is used.

    """
    if logdir is 'synda_logger':
        # Logger initiates by SYNDA worker
        pass
    elif logdir:
        name = os.path.splitext(os.path.basename(os.path.abspath(__file__)))[0]
        logfile = '{0}-{1}-{2}.log'.format(name, datetime.now().strftime("%Y%m%d-%H%M%S"), os.getpid())
        if not os.path.exists(logdir):
            os.mkdir(logdir)
        logging.basicConfig(filename=os.path.join(logdir, logfile),
                            level=logging.DEBUG,
                            format='%(asctime)s %(levelname)s %(message)s',
                            datefmt='%Y/%m/%d %I:%M:%S %p')
    else:
        logging.basicConfig(level=logging.DEBUG,
                            format='%(message)s')


def check_directory(path):
    """
    Checks if the supplied directory exists. The path is normalized before without trailing slash.

    :param list paths: The path to check
    :returns: The same path if exists
    :rtype: *str*
    :raises Error: If the directory does not exist

    """
    directory = os.path.normpath(path)
    if not os.path.isdir(directory):
        raise Exception('No such directory: {0}'.format(directory))
    return directory


def config_parse(config_path):
    """
    Parses the configuration file if exists.

    :param str config_path: The absolute or relative path of the configuration file
    :returns: The configuration file parser
    :rtype: *dict*
    :raises Error: If no configuration file exists
    :raises Error: If the configuration file parsing fails

    """
    if not os.path.isfile(config_path):
        raise Exception('Configuration file not found')
    cfg = ConfigParser.ConfigParser()
    cfg.read(config_path)
    if not cfg:
        raise Exception('Configuration file parsing failed')
    return cfg


def time_init(ctx):
    """
    Returns the required referenced time properties from first file into processing context:
     * The calendar is then used for all files scanned into the variable path,
     * The NetCDF time units attribute has to be unchanged in respect with CF convention and archives designs,
     * The frequency and the version are also read to detect instantaneous time axis.

    :param dict ctx: The processing context (as a :func:`ProcessingContext` class instance)
    :raises Error: If NetCDF time units attribute is missing
    :raises Error: If NetCDF calendar attribute is missing

    """
    file = yield_inputs(ctx).next()[0]
    data = Dataset('{0}/{1}'.format(ctx.directory, file), 'r')
    if 'units' in data.variables['time'].ncattrs():
            ctx.tunits = control_time_units(data.variables['time'].units, ctx)
    else:
        raise Exception('Time units attribute is missing.')
    ctx.funits = convert_time_units(ctx.tunits, ctx.attributes['frequency'])
    if 'calendar' in data.variables['time'].ncattrs():
        ctx.calendar = data.variables['time'].calendar
    else:
        raise Exception('Time calendar attribute is missing.')
    if ctx.calendar == 'standard' and ctx.attributes['model'] == 'CMCC-CM':
        ctx.calendar = 'proleptic_gregorian'
    data.close()
    logging.warning('Version:'.ljust(25)+'{0}'.format(ctx.attributes['version']))
    logging.warning('Frequency:'.ljust(25)+'{0}'.format(ctx.attributes['frequency']))
    logging.warning('Calendar:'.ljust(25)+'{0}'.format(ctx.calendar))
    logging.warning('Time units:'.ljust(25)+'{0}'.format(ctx.tunits))


def control_time_units(tunits, ctx):
    """
    Controls the time units format as at least "days since YYYY-MM-DD".
    The time units can be forced within configuration file using the ``time_units_default`` option.

    :param str tunits: The NetCDF time units string from file
    :param str project: The MIP project
    :returns: The appropriate time units string formatted and controlled depending on the project
    :rtype: *str*

    """
    units = tunits.split()
    units[0] = unicode('days')
    if len(units[2].split('-')) == 1:
        units[2] = units[2] + '-{0}-{1}'.format('01', '01')
    elif len(units[2].split('-')) == 2:
        units[2] = units[2] + '-{0}'.format('01')
    try:
        time_units_default = eval(ctx.cfg.get(ctx.project, 'time_units_default'))
        if ' '.join(units) != time_units_default:
            logging.warning('Invalid time units. Replace "{0}" by "{1}"'.format(' '.join(units), time_units_default))
        return time_units_default
    except:
        return ' '.join(units)


def convert_time_units(unit, frequency):
    """
    Converts default time units from file into time units using the MIP frequency.
    As en example, for a 3-hourly file, the time units "days since YYYY-MM-DD" becomes "hours since YYYY-MM-DD".

    :param str tunits: The NetCDF time units string from file
    :param str frequency: The time frequency
    :returns: The converted time units string
    :rtype: *str*

    """
    units = {'subhr': 'minutes',
             '3hr': 'hours',
             '6hr': 'hours',
             'day': 'days',
             'mon': 'months',
             'yr': 'years'}
    return unit.replace('days', units[frequency])


def counted(fct):
    """
    Decorator used to count all file process calls.

    :param callable fct: The function to monitor
    :returns: A wrapped function with a ``.called`` attribute
    :rtype: *callable*

    """
    @wraps(fct)  # Convenience decorator to keep the file_process docstring
    def wrapper(*args, **kwargs):
        wrapper.called += 1
        return fct(*args, **kwargs)
    wrapper.called = 0
    wrapper.__name__ = fct.__name__
    return wrapper


@counted
def time_axis_processing(inputs):
    """
    time_axis_processing(inputs)

    Time axis process that\:
     * Deduces start and end dates from filename,
     * Rebuilds the theoretical time axis (using frenquency, calendar, etc.),
     * Compares the theoretical time axis with the time axis from the file,
     * Compares the last theoretical date with the end date from the filename,
     * Checks if the expected time units keep unchanged,
     * Checks the squareness and the consistency of time boundaries,
     * Rewrites (with ``-w/--write`` mode) the new time axis,
     * Computes the new checksum if modified,
     * Tracebacks the status.

    :param tuple inputs: A tuple with the file path and the processing context
    :returns: The time axis status as an :func:`AxisStatus` instance
    :rtype: *dict*

    """
    # Extract inputs from tuple
    filename, ctx = inputs
    # Extract start and end dates from filename
    start_date, end_date = dates_from_filename(ctx.project, filename, ctx.calendar, ctx.pattern)
    start = Date2num(start_date, units=ctx.funits, calendar=ctx.calendar)
    # Set time length, True/False instant axis and incrementation in frequency units
    data = Dataset('{0}/{1}'.format(ctx.directory, filename), 'r+')
    length = data.variables['time'].shape[0]
    instant = is_instant_time_axis(ctx)
    inc = time_inc(ctx.attributes['frequency'])
    # Instanciates object to display axis status
    status = AxisStatus()
    status.directory = ctx.directory
    status.file = filename
    status.start = date_print(start_date)
    status.end = date_print(end_date)
    status.steps = length
    status.calendar = ctx.calendar
    status.units = control_time_units(data.variables['time'].units, ctx)
    if instant:
        status.instant = True
    # Rebuild a proper time axis
    axis_hp, last_hp = rebuild_time_axis(start, length, instant, inc, ctx)  # High precision
    axis_lp, last_lp = rebuild_time_axis(trunc(start, 5), length, instant, inc, ctx)  # Low precision avoiding float precision issues
    # Control consistency between last time date and end date from filename
    if not last_date_checker(date_print(last_hp), date_print(end_date)) and not last_date_checker(date_print(last_lp), date_print(end_date)):
        status.control.append('003')
        logging.warning('ERROO3 - {0} - The date from last theoretical time step differs from the end date from filename'.format(filename))
    else:
        if last_date_checker(date_print(last_hp), date_print(end_date)):
            status.last = date_print(last_hp)
            status.axis = axis_hp
        elif last_date_checker(date_print(last_lp), date_print(end_date)):
            status.last = date_print(last_lp)
            status.axis = axis_lp
        # Control consistency between instant time and time boundaries
        if instant and ('time_bnds' in data.variables.keys()):
            status.control.append('004')
            logging.warning('ERROO4 - {0} - An instantaneous time axis should not embeded time boundaries'.format(filename))
            # Delete time bounds and bounds attribute from file
            if ctx.write or ctx.force:
                if 'bounds' in data.variables['time'].ncattrs():
                    del data.variables['time'].bounds
                data.close()
                nc_var_delete(ctx, filename, 'time_bnds')
                data = Dataset('{0}/{1}'.format(ctx.directory, filename), 'r+')
        # Control consistency between averaged time and time boundaries
        if not instant and ('time_bnds' not in data.variables.keys()):
            status.control.append('005')
            logging.warning('ERROO5 - {0} - An averaged time axis should embeded time boundaries'.format(filename))
        # Check time axis squareness
        if ctx.check or ctx.write:
            status.time = data.variables['time'][:]
            if time_checker(status.axis, status.time):
                status.control.append('000')
            else:
                status.control.append('001')
                logging.warning('ERROO1 - {0} - Mistaken time axis over one or several time steps'.format(filename))
            # Rebuild, read and check time boundaries squareness if needed
            if 'time_bnds' in data.variables.keys():
                axis_bnds = rebuild_time_bnds(start, length, instant, inc, ctx)
                time_bnds = data.variables['time_bnds'][:, :]
                if time_checker(axis_bnds, time_bnds):
                    status.bnds = True
                else:
                    status.bnds = False
        # Rewrite time axis depending on checking
        if (ctx.write and not time_checker(status.axis, status.time)) or ctx.force:
            data.variables['time'][:] = status.axis
            # Rewrite time units according to MIP requirements (i.e., same units for all files)
            data.variables['time'].units = ctx.tunits
            # Rewrite time boundaries if needed
            if 'time_bnds' in data.variables.keys():
                data.variables['time_bnds'][:, :] = axis_bnds
        # Control consistency between time units
        if ctx.tunits != status.units:
            status.control.append('002')
            logging.warning('ERROO2 - {0} - Time units must be unchanged for the same dataset.'.format(filename))
    # Close file
    data.close()
    # Compute checksum at the end of all modifications and after closing file
    if (ctx.write or ctx.force) and '000' not in status.control:
        status.checksum = checksum(ctx, filename)
    # Return file status
    return status


def checksum(ctx, filename):
    """
    Does the MD5 or SHA256 checksum by the Shell avoiding Python memory limits.

    :param str file: The full path of a file
    :param dict ctx: The processing context (as a :func:`ProcessingContext` class instance)
    :raises Error: If the checksum fails

    """
    assert (ctx.checksum in ['SHA256', 'MD5']), 'Invalid checksum type: {0} instead of MD5 or SHA256'.format(ctx.checksum)
    try:
        if ctx.checksum == 'SHA256':
            shell = os.popen("sha256sum {0} | awk -F ' ' '{{ print $1 }}'".format('{0}/{1}'.format(ctx.directory, filename)), 'r')
        elif ctx.checksum == 'MD5':
            shell = os.popen("md5sum {0} | awk -F ' ' '{{ print $1 }}'".format('{0}/{1}'.format(ctx.directory, filename)), 'r')
        return shell.readline()[:-1]
    except:
        raise Exception('Checksum failed for {0}'.format('{0}/{1}'.format(ctx.directory, filename)))


def nc_var_delete(ctx, filename, variable):
    """
    Delete a NetCDF variable using NCO operators.
    A unique filename is generated to avoid multithreading errors.
    To overwrite the input file, the source file is dump using the ``cat`` Shell command-line to avoid Python memory limit.

    :param str filename: The filename
    :param dict ctx: The processing context (as a :func:`ProcessingContext` class instance)
    :param str variable: The NetCDF variable to delete
    :raises Error: If the deletion failed

    """
    # Generate unique filename
    tmp = '{0}{1}'.format(str(uuid4()), '.nc')
    try:
        os.popen("ncks -x -O -v {3} {0}/{1} {0}/{2}".format(ctx.directory, filename, tmp, variable), 'r')
        os.popen("cat {0}/{2} > {0}/{1}".format(ctx.directory, filename, tmp, variable), 'r')
        os.remove('{0}/{1}'.format(ctx.directory, tmp))
    except:
        os.remove('{0}/{1}'.format(ctx.directory, tmp))
        raise Exception('Deleting "{0}" failed for {1}'.format(variable, file))


def dates_from_filename(project, filename, calendar, pattern):
    """
    Returns datetime objetcs for start and end dates from the filename.

    :param str filename: The filename
    :param str project: The MIP project
    :param str calendar: The NetCDF calendar attribute
    :param re Object pattern: The filename pattern as a regex (from `re library <https://docs.python.org/2/library/re.html>`_).

    :returns: ``datetime`` instances for start and end dates from the filename
    :rtype: *datetime.datetime*

    """
    dates = []
    for date in pattern.search(filename).groups()[-2:]:
        digits = untroncated_timestamp(date)
        # Convert string digits to %Y-%m-%d %H:%M:%S format
        date_as_since = ''.join([''.join(triple) for triple in zip(digits[::2], digits[1::2], ['', '-', '-', ' ', ':', ':', ':'])])[:-1]
        # Use num2date to create netCDF4 datetime objects
        dates.append(num2date(0.0, units='days since ' + date_as_since, calendar=calendar))
    return dates


def untroncated_timestamp(timestamp):
    """
    Returns proper digits for yearly and monthly truncated timestamps.
    The dates from filename are filled with the 0 digit to reach 14 digits.
    Consequently, yealry dates starts at January 1st and monthly dates starts at first day of the month.

    :param str timestamp: A date string from a filename
    :returns: The filled timestamp
    :rtype: *str*

    """
    if len(timestamp) == 4:
        # Start year at january 1st
        return (timestamp+'0101').ljust(14, '0')
    elif len(timestamp) == 6:
        # Start month at first day
        return (timestamp+'01').ljust(14, '0')
    else:
        return timestamp.ljust(14, '0')


def Num2date(num_axis, units, calendar):
    """
    A wrapper from ``netCDF4.num2date`` able to handle "years since" and "months since" units.
    If time units are not "years since" or "months since", calls usual ``netcdftime.num2date``.

    :param array num_axis: The numerical time axis following units
    :param str units: The proper time units
    :param str calendar: The NetCDF calendar attribute
    :returns: The corresponding date axis
    :rtype: *array*

    """
    # num_axis is the numerical time axis incremented following units (i.e., by years, months, days, etc).
    if not units.split(' ')[0] in ['years', 'months']:
        # If units are not 'years' or 'months since', call usual netcdftime.num2date:
        return num2date(num_axis, units=units, calendar=calendar)
    else:
        # Return to time refenence with 'days since'
        units_as_days = 'days '+' '.join(units.split(' ')[1:])
        # Convert the time refrence 'units_as_days' as datetime object
        start_date = num2date(0.0, units=units_as_days, calendar=calendar)
        # Control num_axis to always get an Numpy array (even with a scalar)
        num_axis_mod = np.atleast_1d(np.array(num_axis))
        if units.split(' ')[0] == 'years':
            # If units are 'years since'
            # Define the number of maximum and minimum years to build a date axis covering the whole 'num_axis' period
            max_years = np.floor(np.max(num_axis_mod)) + 1
            min_years = np.ceil(np.min(num_axis_mod)) - 1
            # Create a date axis with one year that spans the entire period by year
            years_axis = np.array([add_year(start_date, years_to_add) for years_to_add in np.arange(min_years, max_years+2)])
            # Convert rebuilt years axis as 'number of days since'
            year_axis_as_days = date2num(years_axis, units=units_as_days, calendar=calendar)
            # Index of each years
            yind = np.vectorize(np.int)(np.floor(num_axis_mod))
            # Rebuilt num_axis as 'days since' addint the number of days since referenced time with an half-increment (num_axis_mod - yind) = 0 or 0.5
            num_axis_mod_days = (year_axis_as_days[yind - int(min_years)] + (num_axis_mod - yind) * np.diff(year_axis_as_days)[yind - int(min_years)])
        elif units.split(' ')[0] == 'months':
            # If units are 'months since'
            # Define the number of maximum and minimum months to build a date axis covering the whole 'num_axis' period
            max_months = np.floor(np.max(num_axis_mod)) + 1
            min_months = np.ceil(np.min(num_axis_mod)) - 1
            # Create a date axis with one month that spans the entire period by month
            months_axis = np.array([add_month(start_date, months_to_add) for months_to_add in np.arange(min_months, max_months+12)])
            # Convert rebuilt months axis as 'number of days since'
            months_axis_as_days = date2num(months_axis, units=units_as_days, calendar=calendar)
            # Index of each months
            mind = np.vectorize(np.int)(np.floor(num_axis_mod))
            # Rebuilt num_axis as 'days since' addint the number of days since referenced time with an half-increment (num_axis_mod - mind) = 0 or 0.5
            num_axis_mod_days = (months_axis_as_days[mind - int(min_months)] + (num_axis_mod - mind) * np.diff(months_axis_as_days)[mind - int(min_months)])
        # Convert result as date axis
        return num2date(num_axis_mod_days, units=units_as_days, calendar=calendar)


def Date2num(date_axis, units, calendar):
    """
    A wrapper from ``netCDF4.date2num`` able to handle "years since" and "months since" units.
    If time units are not "years since" or "months since" calls usual ``netcdftime.date2num``.

    :param array num_axis: The date axis following units
    :param str units: The proper time units
    :param str calendar: The NetCDF calendar attribute
    :returns: The corresponding numerical time axis
    :rtype: *array*

    """
    # date_axis is the date time axis incremented following units (i.e., by years, months, days, etc).
    if not units.split(' ')[0] in ['years', 'months']:
        # If units are not 'years' or 'months since', call usual netcdftime.date2num:
        return date2num(date_axis, units=units, calendar=calendar)
    else:
        # Return to time refenence with 'days since'
        units_as_days = 'days '+' '.join(units.split(' ')[1:])
        # Convert date axis as number of days since time reference
        days_axis = date2num(date_axis, units=units_as_days, calendar=calendar)
        # Convert the time refrence 'units_as_days' as datetime object
        start_date = num2date(0.0, units=units_as_days, calendar=calendar)
        # Create years axis from input date axis
        years = np.array([date.year for date in np.atleast_1d(np.array(date_axis))])
        if units.split(' ')[0] == 'years':
            # If units are 'years since'
            # Define the number of maximum and minimum years to build a date axis covering the whole 'num_axis' period
            max_years = np.max(years - start_date.year) + 1
            min_years = np.min(years - start_date.year) - 1
            # Create a date axis with one year that spans the entire period by year
            years_axis = np.array([add_year(start_date, yid) for yid in np.arange(min_years, max_years+2)])
            # Convert years axis as number of days since time reference
            years_axis_as_days = date2num(years_axis, units=units_as_days, calendar=calendar)
            # Find closest index for years_axis_as_days in days_axis
            closest_index = np.searchsorted(years_axis_as_days, days_axis)
            # ???
            return min_years + closest_index + (days_axis - years_axis_as_days[closest_index]) / np.diff(years_axis_as_days)[closest_index]
        elif units.split(' ')[0] == 'months':
            # If units are 'months since'
            # Define the number of maximum and minimum months to build a date axis covering the whole 'num_axis' period
            max_months = np.max(12 * (years - start_date.year)) + 1
            min_months = np.min(12 * (years - start_date.year)) - 1
            # Create a date axis with one month that spans the entire period by month
            months_axis = np.array([add_month(start_date, mid) for mid in np.arange(min_months, max_months+12)])
            # Convert months axis as number of days since time reference
            months_axis_as_days = date2num(months_axis, units=units_as_days, calendar=calendar)
            # Find closest index for months_axis_as_days in days_axis
            closest_index = np.searchsorted(months_axis_as_days, days_axis)
            # ???
            return min_months + closest_index + (days_axis - months_axis_as_days[closest_index]) / np.diff(months_axis_as_days)[closest_index]


def add_month(date, months_to_add):
    """
    Finds the next month from date.

    :param datetime date: Accepts datetime or phony datetime from ``netCDF4.num2date``.
    :param int months_to_add: The number of months to add to the date
    :returns: The final date
    :rtype: *datetime*

    """
    date_next = phony_datetime(year=date.year, month=date.month, day=date.day, hour=date.hour, minute=date.minute, second=date.second)
    years_to_add = int((date.month+months_to_add - np.mod(date.month+months_to_add - 1, 12) - 1) / 12)
    new_month = int(np.mod(date.month+months_to_add - 1, 12)) + 1
    date_next = phony_datetime(year=date.year+years_to_add, month=new_month, day=date.day, hour=date.hour, minute=date.minute, second=date.second)
    return date_next


def add_year(date, years_to_add):
    """
    Finds the next year from date.

    :param datetime date: Accepts datetime or phony datetime from ``netCDF4.num2date``.
    :param int years_to_add: The number of years to add to the date
    :returns: The final date
    :rtype: *datetime*

    """
    date_next = phony_datetime(year=date.year+years_to_add, month=date.month, day=date.day, hour=date.hour, minute=date.minute, second=date.second)
    return date_next


def is_instant_time_axis(ctx):
    """
    Deduces from MIP frequency, variable and realm, if the file requires an instantaneous time axis.

    :param dict ctx: The processing context (as a :func:`ProcessingContext` class instance)
    :returns: True if the file requires an instantaneous time axis
    :rtype: *boolean*

    """
    need_instant_time = eval(ctx.cfg.get(ctx.project, 'need_instant_time'))
    if (ctx.attributes['variable'], ctx.attributes['frequency'], ctx.attributes['realm']) in need_instant_time:
        return True
    else:
        return False


def time_inc(frequency):
    """
    Returns the time incrementation depending on the MIP frequency.

    :param str frequency: The MIP frequency
    :returns: The corresponding time raising value
    :rtype: *int*

    """
    inc = {'subhr': 30,
           '3hr': 3,
           '6hr': 6,
           'day': 1,
           'mon': 1,
           'yr': 1}
    return inc[frequency]


def date_print(date):
    """
    Prints date in format: %Y%m%d %H:%M:%s.

    :param datetime date: Accepts datetime or phony datetime objects
    :returns: The corresponding formatted date to print
    :rtype: *str*

    """
    return '{0:04d}-{1:02d}-{2:02d} {3:02d}:{4:02d}:{5:02d}'.format(date.year, date.month, date.day, date.hour, date.minute, date.second)


def rebuild_time_axis(start, length, instant, inc, ctx):
    """
    Rebuilds time axis from date axis, depending on MIP frequency, calendar and instant status.

    :param float date: The numerical date to start (from ``netCDF4.num2date`` or :func:`Date2num`)
    :param int length: The time axis length (i.e., the timesteps number)
    :param boolean instant: The instant status (from :func:`is_instant_time_axis`)
    :param int inc: The time incrementation (from :func:`time_inc`)
    :param dict ctx: The processing context (as a :func:`ProcessingContext` class instance)
    :returns: The corresponding theoretical time axis
    :rtype: *float array*

    """
    date_axis, last = rebuild_date_axis(start, length, instant, inc, ctx)
    axis = Date2num(date_axis, units=ctx.tunits, calendar=ctx.calendar)
    return axis, last


def rebuild_date_axis(start, length, instant, inc, ctx):
    """
    Rebuilds date axis from numerical time axis, depending on MIP frequency, calendar and instant status.

    :param float date: The numerical date to start (from ``netCDF4.date2num`` or :func:`Date2num`)
    :param int length: The time axis length (i.e., the timesteps number)
    :param boolean instant: The instant status (from :func:`is_instant_time_axis`)
    :param int inc: The time incrementation (from :func:`time_inc`)
    :param dict ctx: The processing context (as a :func:`ProcessingContext` class instance)

    :returns: The corresponding theoretical date axis
    :rtype: *datetime array*

    """
    num_axis = np.arange(start=start, stop=start + length * inc, step=inc)
    if ctx.funits.split(' ')[0] in ['years', 'months']:
        last = Num2date(num_axis[-1], units=ctx.funits, calendar=ctx.calendar)[0]
    else:
        last = Num2date(num_axis[-1], units=ctx.funits, calendar=ctx.calendar)
    if not instant and not inc in [3, 6]:  # To solve non-instant [36]hr files
        num_axis += 0.5 * inc
    date_axis = Num2date(num_axis, units=ctx.funits, calendar=ctx.calendar)
    return date_axis, last


def trunc(f, n):
    """
    Truncates a float f to n decimal places before rounding

    :param float f: The number to truncates
    :param int n: Decimal number to place before rounding
    :returns: The corresponding truncated number
    :rtype: *float*

    """
    slen = len('%.*f' % (n, f))
    return float(str(f)[:slen])


def time_checker(right_axis, test_axis):
    """
    Tests the time axis squareness.

    :param array right_axis: The theoretical time axis (from :func:`rebuild_time_axis`),
    :param array test_axis: The time axis to check from NetCDF file,
    :returns: True if both time axis are exactly the same,
    :rtype: *boolean*

    """
    if np.array_equal(right_axis, test_axis):
        return True
    else:
        return False


def rebuild_time_bnds(start, length, instant, inc, ctx):
    """
    Rebuilds time boundaries from the start date, depending on MIP frequency, calendar and instant status.

    :param float date: The numerical date to start (from ``netCDF4.date2num`` or :func:`Date2num`)
    :param int length: The time axis length (i.e., the timesteps number)
    :param boolean instant: The instant status (from :func:`is_instant_time_axis`)
    :param int inc: The time incrementation (from :func:`time_inc`)
    :param dict ctx: The processing context (as a :func:`ProcessingContext` class instance).

    :returns: The corresponding theoretical time boundaries
    :rtype: *[n, 2] array*

    """
    num_axis_bnds = np.column_stack(((np.arange(start=start, stop=start + length * inc, step=inc)),
                                     (np.arange(start=start, stop=start + (length+1) * inc, step=inc)[1:])))
    if not instant and inc in [3, 6]:  # To solve non-instant [36]hr files
        num_axis_bnds -= 0.5 * inc
    date_axis_bnds = Num2date(num_axis_bnds, units=ctx.funits, calendar=ctx.calendar)
    axis_bnds = Date2num(date_axis_bnds, units=ctx.tunits, calendar=ctx.calendar)
    return axis_bnds


def last_date_checker(last, end):
    """
    Checks if last and end date are the same.

    :param float last: The last timesteps of the theoretical time axis (from :func:`rebuild_time_axis`)
    :param float end: The numerical date to end (from ``netCDF4.date2num`` or :func:`Date2num`)
    :returns: True if both dates are exactly the same
    :rtype: *boolean*

    """
    if last != end:
        return False
    else:
        return True


def yield_inputs(ctx):
    """
    Yields all files to process within tuples with the processing context.

    :param dict ctx: The processing context (as a :func:`ProcessingContext` class instance)
    :returns: Attach the processing context to a file to process as an iterator of tuples
    :rtype: *iter*

    """
    for file in sorted(os.listdir(ctx.directory)):
        if not re.match(ctx.pattern, file):
            logging.warning('{0} has invalid filename and was skipped'.format(file))
            continue
        yield file, ctx


def wrapper(inputs):
    """
    Transparent wrapper for pool map.

    :param tuple inputs: A tuple with the file path and the processing context
    :returns: The :func:`time_axis_processing` call
    :rtype: *callable*
    :raises Error: When a thread-process failed preserving its traceback

    """
    try:
        return time_axis_processing(inputs)
    except:
        logging.exception('A thread-process fails:')


def run(job=None):
    """
    Main process that\:
     * Instanciates processing context,
     * Defines the referenced time properties,
     * Instanciates threads pools,
     * Prints or logs the time axis diagnostics.

    :param dict job: An optionnal dictionnary to supply instead of classical command-line use.

    """
    # Instanciate processing context from command-line arguments or SYNDA job dictionnary
    ctx = ProcessingContext(get_args(job))
    if ctx.attributes['frequency'] == 'fx':
        logging.warning('Skipped "{0}"" frequency because no time axis'.format(ctx.attributes['frequency']))
    else:
        logging.info('Time diagnostic started for {0}'.format(ctx.directory))
        # Set driving time properties (calendar, frequency and time units) from first file in directory
        time_init(ctx)
        logging.info('Files to process:'.ljust(25)+'{0}'.format(len(glob('{0}/*.nc'.format(ctx.directory)))))
        # Process
        pool = ThreadPool(int(ctx.cfg.defaults()['threads_number']))
        outputs = pool.imap(wrapper, yield_inputs(ctx))
        # Returns diagnostic for each file
        if not job:
            job = {}
        job['files'] = {}
        for output in outputs:
            logging.info('==> Filename:'.ljust(25)+'{0}'.format(output.file))
            if ctx.verbose:
                logging.info('-> Start:'.ljust(25)+'{0}'.format(str(output.start)))
                logging.info('-> End:'.ljust(25)+'{0}'.format(str(output.end)))
                logging.info('-> Last:'.ljust(25)+'{0}'.format(str(output.last)))
            logging.info('-> Timesteps:'.ljust(25)+'{0}'.format(output.steps))
            logging.info('-> Instant time:'.ljust(25)+'{0}'.format(output.instant))
            logging.info('-> Time axis status:'.ljust(25)+'{0}'.format(','.join(output.control)))
            logging.info('-> Time boundaries:'.ljust(25)+'{0}'.format(output.bnds))
            logging.info('-> New checksum:'.ljust(25)+'{0}'.format(output.checksum))
            if ctx.verbose:
                logging.info('-> Time axis:')
                logging.info('{0}'.format(fill(' | '.join(map(str, output.time.tolist())), width=100)))
                logging.info('-> Theoretical axis:')
                logging.info('{0}'.format(fill(' | '.join(map(str, output.axis.tolist())), width=100)))
            # Return diagnostic to SYNDA using job dictionnary
            job['files'][output.file] = {}
            job['files'][output.file]['version'] = ctx.attributes['version']
            job['files'][output.file]['calendar'] = output.calendar
            job['files'][output.file]['start'] = output.start
            job['files'][output.file]['end'] = output.end
            job['files'][output.file]['last'] = output.last
            job['files'][output.file]['length'] = output.steps
            job['files'][output.file]['instant'] = output.instant
            job['files'][output.file]['bnds'] = output.bnds
            job['files'][output.file]['status'] = ','.join(output.control)
            job['files'][output.file]['new_checksum'] = output.checksum
        # Close tread pool
        pool.close()
        pool.join()
        logging.info('Time diagnostic completed ({0} files scanned)'.format(time_axis_processing.called))
        return job


# Main entry point for stand-alone call.
if __name__ == "__main__":
    run()



#External:
import numpy as np
import math
import time
import h5netcdf.legacyapi as netCDF4_h5
import netCDF4
import h5py
import datetime
import copy
import os
from collections import OrderedDict
#import dask.array as da

#Internal:
from . import indices_utils

def check_if_opens(dataset,default=False):
    if default: return False
    return True

def get_year_axis(dataset,default=False):
    if default: return np.array([]),np.array([])

    dimensions_list = dataset.dimensions.keys()
    time_dim = find_time_dim(dataset)
    date_axis = get_date_axis(dataset, time_dim)
    year_axis = np.array([date.year for date in date_axis])
    month_axis = np.array([date.month for date in date_axis])
    return year_axis, month_axis

def get_date_axis(dataset, time_dim, default=False):
    if default: return np.array([])
    
    #Use np.asscalar(np.asarray(x)) to ensure that attributes are not arrays if lenght-1
    units = getncattr(dataset.variables[time_dim], 'units')
    if 'calendar' in dataset.variables[time_dim].ncattrs():
        calendar = getncattr(dataset.variables[time_dim], 'calendar')
    else:
        calendar = None
    return get_date_axis_from_units_and_calendar(dataset.variables[time_dim][:], units, calendar)

def get_date_axis_from_units_and_calendar(time_axis, units, calendar, default=False):
    if default: return np.array([])

    if units == 'day as %Y%m%d.%f':
        date_axis = get_date_axis_absolute(time_axis)
    else:
        date_axis = get_date_axis_relative(time_axis, units, calendar)
    return date_axis

def get_date_axis_relative(time_axis,units,calendar,default=False):
    if default: return np.array([])
    if calendar is not None:
        try:
            date_axis = netCDF4.num2date(time_axis,units=units,calendar=calendar)
        except ValueError:
            if (
                (units=='days since 0-01-01 00:00:00' and
                calendar=='365_day') or
                (units=='days since 0-1-1 00:00:00' and
                calendar=='365_day') 
                ):
                date_axis = netCDF4.num2date(time_axis-365.0,units='days since 1-01-01 00:00:00',calendar=calendar)
            else:
                raise
    else:
        date_axis = netCDF4.num2date(time_axis,units=units)
    return date_axis

def get_date_axis_absolute(time_axis,default=False):
    if default: return np.array([])
    return map(convert_to_date_absolute,time_axis)

def get_time(dataset,time_var='time',default=False):
    if default: return np.array([])
    time_dim = find_time_dim(dataset,time_var=time_var)
    time_axis, attributes = retrieve_dimension(dataset, time_dim)
    date_axis = create_date_axis_from_time_axis(time_axis, attributes)
    return date_axis

def get_time_axis_relative(date_axis, units, calendar, default=False):
    if default: return np.array([])
    if calendar != None:
        try:
            time_axis = netCDF4.date2num(date_axis, units=units, calendar=calendar)
        except ValueError:
            if (
                ( units == 'days since 0-01-01 00:00:00' and
                  calendar == '365_day') or
                ( units == 'days since 0-1-1 00:00:00' and
                  calendar == '365_day') 
                ):
                time_axis = netCDF4.date2num(date_axis, units='days since 1-01-01 00:00:00', calendar=calendar) + 365.0
            else:
                raise
    else:
        time_axis = netCDF4.date2num(date_axis, units=units)
    return time_axis

def convert_to_date_absolute(absolute_time):
    year=int(math.floor(absolute_time/1e4))
    remainder=absolute_time-year*1e4
    month=int(math.floor(remainder/1e2))
    remainder-=month*1e2
    day=int(math.floor(remainder))
    remainder-=day
    remainder*=24.0
    hour=int(math.floor(remainder))
    remainder-=hour
    remainder*=60.0
    minute=int(math.floor(remainder))
    remainder-=minute
    remainder*=60.0
    seconds=int(math.floor(remainder))
    return datetime.datetime(year,month,day,hour,minute,seconds)

def replicate_full_netcdf_recursive(dataset,output,transform=(lambda x,y,z:y),slices=dict(),
                                                check_empty=False,default=False):
    if default: return output

    for var_name in dataset.variables.keys():
        replicate_and_copy_variable(dataset,output,var_name,transform=transform,slices=slices,check_empty=check_empty)
    if len(dataset.groups.keys())>0:
        for group in dataset.groups.keys():
            output_grp = replicate_group(dataset,output,group)
            replicate_full_netcdf_recursive(dataset.groups[group],output_grp,transform=transform,slices=slices,check_empty=check_empty)
    return output

def dimension_compatibility(dataset,output,dim,default=False):
    if default: return False

    if (dim in output.dimensions.keys()
        and _dim_len(output,dim)!=_dim_len(dataset,dim)):
        #Dimensions mismatch, return without writing anything
        return False
    elif ( (dim in dataset.variables.keys() and
          dim in output.variables.keys()) and
          ( len(output.variables[dim]) != len(dataset.variables[dim]) or 
           (dataset.variables[dim][:] != dataset.variables[dim][:]).any())):
        #Dimensions variables mismatch, return without writing anything
        return False
    else:
        return True

def check_dimensions_compatibility(dataset,output,var_name,exclude_unlimited=False,default=False):
    if default: return False
    for dim in dataset.variables[var_name].dimensions:
        #The dimensions might be in the parent group:
        if not dim in dataset.dimensions.keys():
            dataset_parent = dataset.parent
        elif not dim in dataset.variables.keys():
            #Important check for h5netcdf
            dataset_parent = dataset.parent
        else:
            dataset_parent = dataset

        if not dim in output.dimensions.keys():
            output_parent = output.parent
        else:
            output_parent = output

        if not _isunlimited(dataset_parent,dim) or not exclude_unlimited:
            if not dimension_compatibility(dataset_parent,output_parent,dim):
                return False
    return True

def _isunlimited(dataset,dim):
    if (isinstance(dataset, netCDF4_h5.Dataset) or
        isinstance(dataset, netCDF4_h5.Group)):
        var_list_with_dim=[var for var in dataset.variables.keys() 
                                if dim in dataset.variables[var].dimensions]
        if len(var_list_with_dim)==0:
            return False

        if np.all([dataset._h5group[var].maxshape[
                                    list(dataset.variables[var].dimensions).index(dim)]==None
                                for var in var_list_with_dim]):
            #If the maxshape of dimension for all variables with dimenion is None, it is unlimited!
            return True
        else:
            return False
    else:
        return dataset.dimensions[dim].isunlimited()

def _dim_len(dataset,dim):
    if (isinstance(dataset, netCDF4_h5.Dataset) or
        isinstance(dataset, netCDF4_h5.Group)):
        return dataset.dimensions[dim]
    else:
        return len(dataset.dimensions[dim])

def _datatype(dataset,var):
    if (isinstance(dataset, netCDF4_h5.Dataset) or
        isinstance(dataset, netCDF4_h5.Group)):
        dtype = dataset.variables[var].dtype
        #if dtype == 'object':
        #    return np.dtype(str)
        #else:
        return np.dtype(dtype)
    else:
        return dataset.variables[var].datatype

def append_record(dataset, output, default=False):
    record_dimensions = dict()
    if default: return record_dimensions
    for dim in dataset.dimensions.keys():
        if (     dim in dataset.variables.keys()
             and dim in output.dimensions.keys()
             and dim in output.variables.keys()
             and _isunlimited(dataset,dim)):
             append_slice = ensure_compatible_time_units(dataset, output, dim)
             record_dimensions[dim] = {'append_slice': append_slice}
    return record_dimensions

def ensure_compatible_time_units(dataset, output, dim, default=False):
    if default: return dataset.variables[dim][:]
    try:
        units = dict()
        calendar = dict()
        for desc, data in [ ('source', dataset), ('dest', output) ]:
            units[desc] = getncattr(data.variables[dim], 'units')
            if 'calendar' in data.variables[dim].ncattrs():
                calendar[desc] = getncattr(data.variables[dim], 'calendar')
            else:
                calendar[desc] = 'standard'

        converted_dim = netCDF4.date2num(netCDF4.num2date(dataset.variables[dim][:],
                                                          units['source'], calendar=calendar['source']),
                                         units['dest'],calendar=calendar['dest'])

        dest_dim = output.variables[dim][:]
    except KeyError:
        converted_dim = dataset.variables[dim][:]
        dest_dim = output.variables[dim][:]

    overlapping_source_mask = np.in1d(converted_dim, dest_dim)
    if np.any(overlapping_source_mask):
        non_overlapping_source_mask = np.invert(overlapping_source_mask)
        if sum(non_overlapping_source_mask) > 0:
            append_slice = slice(len(dest_dim),len(dest_dim)+
                                               sum(non_overlapping_source_mask),1)
            output.variables[dim][append_slice] = converted_dim[non_overlapping_source_mask]

            dest_dim = output.variables[dim][:]

        sorting_dest_dim = np.argsort(dest_dim)
        append_indices_or_slice = sorting_dest_dim[np.searchsorted(dest_dim, converted_dim,
                                                                   sorter=sorting_dest_dim)]
    else:
        append_indices_or_slice = slice(len(dest_dim),len(dest_dim)+
                                                  len(converted_dim),1)
        output.variables[dim][append_indices_or_slice] = converted_dim
    return append_indices_or_slice

def append_and_copy_variable(dataset,output,var_name,record_dimensions,
                             datatype=None,fill_value=None,add_dim=None,
                             chunksize=None,zlib=False,check_empty=False,default=False):
    if default: return output

    if len(set(record_dimensions.keys()).intersection(dataset.variables[var_name].dimensions))==0:
        #Variable does not contain a record dimension, return
        return output
   
    variable_size = min(dataset.variables[var_name].shape)
    storage_size = variable_size
    if '_h5ds' in dir(dataset):
        #Use the hdf5 library to find the real size of the stored array:
        variable_size = dataset.variables[var_name]._h5ds.size
        storage_size = dataset.variables[var_name]._h5ds.id.get_storage_size()

    if variable_size > 0 and storage_size > 0:
        max_request = 450.0 #maximum request in Mb
        #use dask array store
        max_first_dim_steps = max(
                        int(np.floor(max_request*1024*1024/(32*np.prod(dataset.variables[var_name].shape[1:])))),
                        1)

    #Using dask. Not working yet:
    #    source = da.from_array(dataset.variables[var_name], 
    #                           chunks=(max_first_dim_steps,)+dataset.variables[var_name].shape[1:])
    #
    #    getitem_tuple = tuple([ slice(0,_dim_len(dataset,dim),1) if not dim in record_dimensions.keys()
    #                                                      else record_dimensions[dim]['append_slice']
    #                                                      for dim in dataset.variables[var_name].dimensions ])
    #    dest = ( da.from_array(output.variables[var_name], 
    #                         chunks=(1,)+output.variables[var_name].shape[1:])[getitem_tuple]
    #                         .rechunk((max_first_dim_steps,)+dataset.variables[var_name].shape[1:]) )
    #
    #    da.store(source, dest)
    #return output

        num_first_dim_chunk = int(np.ceil(dataset.variables[var_name].shape[0]/float(max_first_dim_steps)))
        for first_dim_chunk in range(num_first_dim_chunk):
            first_dim_slice = slice(first_dim_chunk*max_first_dim_steps,
                             min((first_dim_chunk+1)*max_first_dim_steps,dataset.variables[var_name].shape[0])
                             ,1)
            output = append_dataset_first_dim_slice(dataset,output,var_name,first_dim_slice,record_dimensions,check_empty)
    return output

def append_dataset_first_dim_slice(dataset,output,var_name,first_dim_slice,record_dimensions,check_empty):
    #Create a setitem tuple
    setitem_list = [ slice(0,_dim_len(dataset,dim),1) if not dim in record_dimensions.keys()
                                                               else record_dimensions[dim]['append_slice']
                                                              for dim in dataset.variables[var_name].dimensions]
    #Pick a first_dim_slice along the first dimension:
    setitem_list[0] = indices_utils.slice_a_slice(setitem_list[0], first_dim_slice)
    temp = dataset.variables[var_name][first_dim_slice, ...]
    assign_not_masked(temp, output.variables[var_name], setitem_list, check_empty)
    return output

def assign_not_masked(source, dest, setitem_list, check_empty):
    #Assign only if not masked everywhere:
    if ( not 'mask' in dir(source) or 
         not check_empty or 
         not source.mask.all() ):
            
        try:
            dest[tuple(setitem_list)] = np.ma.filled(source)
        except AttributeError as e:
            errors_to_ignore = ["'str' object has no attribute 'size'",
                                "'unicode' object has no attribute 'size'"]
            if (unicode(e) in errors_to_ignore
                and len(setitem_list) == 1):
                for source_id, dest_id in enumerate(setitem_list[0]):
                    dest[dest_id] = source[source_id]
            else:
                raise
    return

def replicate_and_copy_variable(dataset,output,var_name,
                                datatype=None,fill_value=None,
                                add_dim=None,
                                chunksize=None,zlib=False,
                                transform=(lambda x,y,z:y),
                                slices=dict(),
                                check_empty=False,default=False):

    if default: return output

    if not isinstance(slices,dict):
        #assume it is a function that takes the dataset as input and outputs
        #a slicing dict
        comp_slices = slices(dataset)
    else:
        comp_slices = slices

    replicate_netcdf_var(dataset, output, var_name,
                        datatype=datatype, fill_value=fill_value,
                        add_dim=add_dim,
                        slices=comp_slices,
                        chunksize=chunksize, zlib=zlib)

    #Apply a transformation if dimensions are in slices:
    if set(comp_slices.keys()).issubset(dataset.variables[var_name].dimensions):
        transform(dataset, output, comp_slices)

    if len(dataset.variables[var_name].dimensions) == 0:
        #scalar variable:
        #try:
        value = dataset.variables[var_name][...]
        if not np.ma.is_masked(value):
            #if not masked, assign. Otherwise, do nothing
            try:
                output.variables[var_name][...] = value
            except AttributeError as e:
                # This appears to be a netcdf4 bug. Skip this error at moment.
                if not (unicode(e) == "type object 'str' has no attribute 'kind'" and
                        value == ''):
                    raise
        #except IOError as e:
        #    # Loading scalar in h5py is not stable from version to version at the moment:
        #    errors_to_ignore = ["Can't read data (No appropriate function for conversion path)"]
        #    if not unicode(e) in errors_to_ignore:
        #        raise
        return output

    variable_size = min(dataset.variables[var_name].shape)
    storage_size = variable_size
    if '_h5ds' in dir(dataset):
        #Use the hdf5 library to find the real size of the stored array:
        variable_size = dataset.variables[var_name]._h5ds.size
        storage_size = dataset.variables[var_name]._h5ds.id.get_storage_size()

    if variable_size > 0 and storage_size > 0:
        max_request = 450.0 #maximum request in Mb
        #max_request=9000.0 #maximum request in Mb

        #Create the output variable shape, allowing slices:
        var_shape = tuple([ dataset.variables[var_name].shape[dim_id] if not dim in comp_slices.keys()
                            else len(np.arange(dataset.variables[var_name].shape[dim_id])[comp_slices[dim]])
                            for dim_id,dim in enumerate(dataset.variables[var_name].dimensions) ])
        max_first_dim_steps = max(
                        int(np.floor(max_request*1024*1024/(32*np.prod(var_shape[1:])))),
                        1)

#    # Using dask. Not working yet:
#        getitem_tuple = tuple([ comp_slices[var_dim] if var_dim in comp_slices.keys()
#                                else slice(None,None,1) for var_dim in
#                                dataset.variables[var_name].dimensions ])
#        source = ( da.from_array(dataset.variables[var_name], 
#                               chunks=(1,)+dataset.variables[var_name].shape[1:])[getitem_tuple]
#                               .rechunk((max_first_dim_steps,)+output.variables[var_name].shape[1:]) )
#        dest = ( da.from_array(output.variables[var_name], 
#                               chunks=(max_first_dim_steps,)+output.variables[var_name].shape[1:]))
#    
#        print(dest.dask)
#        da.store(source, dest)
#    #    da.store(source, output.variables[var_name])
#    return output

#def dummy():
        num_first_dim_chunk = int(np.ceil(var_shape[0]/float(max_first_dim_steps)))

        for first_dim_chunk in range(num_first_dim_chunk):
            first_dim_slice = slice(first_dim_chunk*max_first_dim_steps,
                                    min((first_dim_chunk+1)*max_first_dim_steps,var_shape[0])
                                    ,1)
            output = copy_dataset_first_dim_slice(dataset, output, var_name, first_dim_slice,
                                                  check_empty,slices=comp_slices)
#        return output
    return output

def copy_dataset_first_dim_slice(dataset, output, var_name, first_dim_slice, check_empty, slices=dict()):
    combined_slices = slices.copy()
    first_dim = dataset.variables[var_name].dimensions[0]
    if first_dim in combined_slices:
        combined_slices[first_dim] = indices_utils.slice_a_slice(combined_slices[first_dim], first_dim_slice)
    else:
        combined_slices[first_dim] = first_dim_slice
                
    getitem_tuple = tuple([ combined_slices[var_dim] if var_dim in combined_slices.keys()
                            else slice(None,None,1) for var_dim in
                            dataset.variables[var_name].dimensions ])

    temp = dataset.variables[var_name][getitem_tuple]
    assign_not_masked(temp, output.variables[var_name], [first_dim_slice, Ellipsis], check_empty)
    return output

def replicate_group(dataset, output, group_name, default=False):
    if default: return output
    output_grp = create_group(dataset, output, group_name)
    replicate_netcdf_file(dataset.groups[group_name], output_grp)
    return output_grp

def create_group(dataset, output, group_name, default=False):
    if default: return output
    if not group_name in output.groups.keys():
        output_grp = output.createGroup(group_name)
    else:
        output_grp = output.groups[group_name]
    return output_grp
    
def replicate_netcdf_file(dataset, output, default=False):
    if default: return output

    for att in dataset.ncattrs():
        #Use np.asscalar(np.asarray()) for backward and forward compatibility:
        att_val = getncattr(dataset, att)
        
        #This fix is for compatitbility with h5netcdf:
        if ( 'dtype' in dir(att_val) and
            att_val.dtype == np.dtype('O')):
            if len(att_val) == 1:
                att_val = att_val[0]
            else:
                att_val = np.asarray(att_val, dtype='str')

        if 'encode' in dir(att_val):
            try:
                att_val = str(att_val.encode('ascii', 'replace'))
            except UnicodeDecodeError as e:
                att_var = str(att_val)

            
        if (not att in output.ncattrs() and
            att != 'cdb_query_temp'):
            setncattr(output, att, att_val)
    return output

def setncattr(output, att, att_val):
    try:
        output.setncattr_string(att, att_val)
    except AttributeError as e:
        output.setncattr(att, att_val)
    return

def getncattr(dataset, att):
    return _toscalar(np.asarray(dataset.getncattr(att)))

def _is_dimension_present(dataset,dim):
    if dim in dataset.dimensions:
        return True
    elif dataset.parent is not None:
        return _is_dimension_present(dataset.parent, dim)
    else:
        return False

def replicate_netcdf_var_dimensions(dataset, output, var,
                                    slices=dict(),
                                    datatype=None,
                                    fill_value=None,
                                    add_dim=None,
                                    chunksize=None,zlib=False,default=False):
    if default: return output
    for dims in dataset.variables[var].dimensions:
        if ( not _is_dimension_present(output, dims) and 
             _is_dimension_present(dataset, dims) ):
            if _isunlimited(dataset, dims):
                output.createDimension(dims, None)
            elif dims in slices.keys():
                output.createDimension(dims, len(np.arange(_dim_len(dataset, dims))[slices[dims]]))
            else:
                output.createDimension(dims, _dim_len(dataset, dims))
            if dims in dataset.variables:
                replicate_netcdf_var(dataset, output, dims, zlib=True, slices=slices)
                if dims in slices.keys():
                    output.variables[dims][:] = dataset.variables[dims][slices[dims]]
                else:
                    output.variables[dims][:] = dataset.variables[dims][:]
                if ('bounds' in output.variables[dims].ncattrs() and
                    getncattr(output.variables[dims], 'bounds')
                    in dataset.variables.keys()):
                    var_bounds = getncattr(output.variables[dims], 'bounds')
                    if not var_bounds in output.variables.keys():
                        output=replicate_netcdf_var(dataset, output, var_bounds,
                                                    zlib=True, slices=slices)
                        if dims in slices.keys():
                            getitem_tuple = tuple([ slices[var_bounds_dim] if var_bounds_dim in slices.keys()
                                                  else slice(None,None,1) for var_bounds_dim in
                                                  dataset.variables[var_bounds].dimensions ])
                            output.variables[var_bounds][:] = dataset.variables[var_bounds][getitem_tuple]
                        else:
                            output.variables[var_bounds][:] = dataset.variables[var_bounds][:]
            else:
                #Create a dummy dimension variable:
                dim_var = output.createVariable(dims,np.float,(dims,),chunksizes=(1,))
                if dims in slices.keys():
                    dim_var[:]=np.arange(_dim_len(dataset,dims))[slices[dims]]
                else:
                    dim_var[:]=np.arange(_dim_len(dataset,dims))
    return output

def replicate_netcdf_other_var(dataset,output,var,time_dim,default=False):
    if default: return output
    #Replicates all variables except specified variable:
    variables_list=[ other_var for other_var in variables_list_with_time_dim(dataset,time_dim)
                                if other_var!=var]
    for other_var in variables_list:
        output=replicate_netcdf_var(dataset,output,other_var)
    return output

def replicate_netcdf_var(dataset,output,var,
                        slices=dict(),
                        datatype=None,fill_value=None,add_dim=None,chunksize=None,zlib=False,default=False):
    if default: 
        #Create empty variable:
        output.createVariable(var,'d',())
        return output

    if not var in dataset.variables.keys():
        return output

    output=replicate_netcdf_var_dimensions(dataset,output,var,slices=slices)
    if var in output.variables.keys():
        #var is a dimension variable and does not need to be created:
        return output

    if datatype==None: datatype=_datatype(dataset,var)
    if (isinstance(datatype,netCDF4.CompoundType) and
        not datatype.name in output.cmptypes.keys()):
        datatype=output.createCompoundType(datatype.dtype,datatype.name)

    #Weird fix for strings:
    #if 'str' in dir(datatype) and 'S1' in datatype.str:
    #    datatype='S1'

    kwargs=dict()
    if (fill_value==None and 
        '_FillValue' in dataset.variables[var].ncattrs() and 
        datatype==_datatype(dataset,var)):
            kwargs['fill_value'] = getncattr(dataset.variables[var], '_FillValue')
    else:
        kwargs['fill_value'] = fill_value

    if not zlib:
        if dataset.variables[var].filters()==None:
            kwargs['zlib']=False
        else:
            for item in dataset.variables[var].filters():
                kwargs[item]=dataset.variables[var].filters()[item]
    else:
        kwargs['zlib']=zlib
    
    if var not in output.variables.keys():
        dimensions=dataset.variables[var].dimensions
        time_dim=find_time_dim(dataset)
        if add_dim:
            dimensions+=(add_dim,)
        var_shape=tuple([dataset.variables[var].shape[dim_id] if not dim in slices.keys()
                                               else len(np.arange(dataset.variables[var].shape[dim_id])[slices[dim]])
                                               for dim_id,dim in enumerate(dimensions)])
        if chunksize==-1:
            chunksizes=tuple([1 if dim==time_dim else var_shape[dim_id] for dim_id,dim in enumerate(dimensions)])
        elif dataset.variables[var].chunking()=='contiguous':
            if kwargs['zlib']:
                chunksizes=tuple([1 if dim==time_dim else var_shape[dim_id] for dim_id,dim in enumerate(dimensions)])
            else:
                chunksizes=tuple([1 for dim_id,dim in enumerate(dimensions)])
        else:
            if len(set(dimensions).intersection(slices.keys()))>0:
                if kwargs['zlib']:
                    chunksizes=tuple([1 if dim==time_dim else var_shape[dim_id] for dim_id,dim in enumerate(dimensions)])
                else:
                    chunksizes=tuple([1 for dim_id,dim in enumerate(dimensions)])
            else:
                chunksizes=dataset.variables[var].chunking()
        kwargs['chunksizes']=chunksizes
        out_var=output.createVariable(var,datatype,dimensions,**kwargs)
    output = replicate_netcdf_var_att(dataset,output,var)
    return output
    #return out_var

def _toscalar(x):
    try:
        return np.asscalar(x)
    except (AttributeError, ValueError) as e:
        return x

def replicate_netcdf_var_att(dataset,output,var,default=False):
    if default: return output
    for att in dataset.variables[var].ncattrs():
        #Use np.asscalar(np.asarray()) for backward and forward compatibility:
        att_val = getncattr(dataset.variables[var], att)
        if isinstance(att_val,dict):
            atts_pairs=[(att+'.'+key,att_val[key]) for key in att_val.keys()]
        else:
            atts_pairs=[(att,att_val)]
        for att_pair in atts_pairs:
            if att_pair[0][0]!='_':
                if 'encode' in dir(att_pair[1]):
                    att_val=att_pair[1].encode('ascii','replace')
                else:
                    att_val=att_pair[1]
                if 'encode' in dir(att_pair[0]):
                    att=att_pair[0].encode('ascii','replace')
                else:
                    att=att_pair[0]
                setncattr(output.variables[var],att,att_val)
    return output

def create_time_axis(dataset,output,time_axis,time_var='time',default=False):
    if default: return output
    #output.createDimension(time_dim,len(time_axis))
    time_dim=find_time_dim(dataset,time_var=time_var)
    output.createDimension(time_dim,None)
    time = output.createVariable(time_dim,'d',(time_dim,),chunksizes=(1,))
    if dataset == None:
        setncattr(time, 'calendar', 'standard')
        setncattr(time, 'units', 'days since '+str(time_axis[0]))
    else:
        setncattr(time, 'calendar', netcdf_calendar(dataset, time_var=time_var))
        time_var = find_time_var(dataset, time_var=time_var)
        #Use np.asscalar(np.asarray()) for backward and forward compatibility:
        setncattr(time, 'units', str(getncattr(dataset.variables[time_var], 'units')))
    time[:] = time_axis
    return output

def create_time_axis_date(output, time_axis, units, calendar, time_dim='time'):
    output.createDimension(time_dim, None)
    time = output.createVariable(time_dim, 'd', (time_dim,), chunksizes=(1,))
    setncattr(time, 'calendar', calendar)
    setncattr(time, 'units', units)
    time[:] = get_time_axis_relative(time_axis, getncattr(time, 'units'),
                                                getncattr(time, 'calendar') )
    return

def netcdf_calendar(dataset, time_var='time', default=False):
    calendar = 'standard'
    if default: return calendar

    time_var = find_time_var(dataset, time_var=time_var)
    if time_var is not None:
        if 'calendar' in dataset.variables[time_var].ncattrs():
            #Use np.asscalar(np.asarray()) for backward and forward compatibility:
            calendar = getncattr(dataset.variables[time_var], 'calendar')
        if 'encode' in dir(calendar):
            calendar = calendar.encode('ascii', 'replace')
    return calendar
    
def find_time_var(dataset, time_var='time', default=False):
    if default: return time_var
    var_list = dataset.variables.keys()
    return find_time_name_from_list(var_list, time_var)

def find_time_dim(dataset, time_var='time', default=False):
    if default: return time_var
    dim_list = dataset.dimensions.keys()
    return find_time_name_from_list(dim_list, time_var)

def find_time_name_from_list(list_of_names, time_var):
    try:
        return list_of_names[ next(i for i,v in enumerate(list_of_names) 
                              if v.lower() == time_var) ]
    except StopIteration:
        return None

def variables_list_with_time_dim(dataset, time_dim, default=False):
    if default: return []
    return [ var for var in dataset.variables.keys()
             if time_dim in dataset.variables[var].dimensions ]

def find_dimension_type(dataset, time_var='time', default=False):
    dimension_type = OrderedDict()
    if default: return dimension_type

    time_dim = find_time_name_from_list(dataset.dimensions.keys(), time_var)
    for dim in dataset.dimensions.keys():
        if dim != time_dim:
            dimension_type[dim] = _dim_len(dataset,dim)
    return dimension_type

def netcdf_time_units(dataset, time_var='time', default=False):
    units = None
    if default: return units
    time_var = find_time_var(dataset, time_var=time_var)
    if 'units' in dataset.variables[time_var].ncattrs():
        #Use np.asscalar(np.asarray()) for backward and forward compatibility:
        units = getncattr(dataset.variables[time_var], 'units')
    return units

def retrieve_dimension(dataset, dimension, default=False):
    attributes = dict()
    dimension_dataset = np.array([])
    if default: return dimension_dataset, attributes

    if dimension in dataset.variables.keys():
        #Retrieve attributes:
        for att in dataset.variables[dimension].ncattrs():
            #Use np.asscalar(np.asarray()) for backward and forward compatibility:
            attributes[att] = getncattr(dataset.variables[dimension], att)
        #If dimension is available, retrieve
        dimension_dataset = dataset.variables[dimension][...]
    else:
        #If dimension is not avaiable, create a simple indexing dimension
        dimension_dataset = np.arange(_dim_len(dataset, dimension))
    return dimension_dataset, attributes

def retrieve_dimension_list(dataset, var, default=False):
    dimensions = tuple()
    if default: return dimensions
    return dataset.variables[var].dimensions

def retrieve_dimensions_no_time(dataset, var, time_var='time', default=False):
    dimensions_data = dict()
    attributes = dict()
    if default: return dimensions_data, attributes
    dimensions = retrieve_dimension_list(dataset, var)
    time_dim = find_time_name_from_list(dimensions, time_var)
    for dim in dimensions:
        if dim != time_dim:
            dimensions_data[dim], attributes[dim] = retrieve_dimension(dataset, dim)
    return dimensions_data, attributes

def retrieve_variables(dataset, output, zlib=True, default=False):
    if default: return output
    for var_name in dataset.variables.keys():
        output = replicate_and_copy_variable(dataset, output, var_name,
                                             zlib=zlib, check_empty=False)
    return output

def retrieve_variables_no_time(dataset, output, time_dim, zlib=False, default=False):
    if default: return output
    for var in dataset.variables.keys():
        if ( (not time_dim in dataset.variables[var].dimensions) and 
             (not var in output.variables.keys())):
            replicate_and_copy_variable(dataset, output, var, zlib=zlib)
    return output

def find_time_dim_and_replicate_netcdf_file(dataset, output, time_var='time', default=False):
    if default: return ( find_time_dim(dataset, time_var=time_var, default=True),
                         replicate_netcdf_file(dataset, output, default=True) )
    return ( find_time_dim(dataset, time_var=time_var),
             replicate_netcdf_file(dataset, output) )

def create_date_axis_from_time_axis(time_axis, attributes_dict, default=False):
    if default: return np.array([])

    calendar='standard'
    units = attributes_dict['units']
    if 'calendar' in attributes_dict.keys(): 
        calendar = attributes_dict['calendar']

    if units == 'day as %Y%m%d.%f':
        date_axis = np.array(map(convert_to_date_absolute,
                                 native_time_axis))
    else:
        try:
            #Put cmip5_rewrite_time_axis here:
            date_axis = get_date_axis_relative(time_axis, units, calendar)
        except TypeError:
            date_axis = np.array([]) 
    return date_axis

def retrieve_container(dataset, var, dimensions, unsort_dimensions,
                       sort_table, max_request, time_var='time',
                       file_name='', default=False):
    if default: return np.array([])
    remote_dimensions, attributes = retrieve_dimensions_no_time(dataset, var, time_var=time_var)

    indices = copy.copy(dimensions)
    unsort_indices = copy.copy(unsort_dimensions)
    for dim in remote_dimensions.keys():
        indices[dim], unsort_indices[dim] = indices_utils.prepare_indices(
                                            indices_utils.get_indices_from_dim(remote_dimensions[dim],
                                                                               indices[dim]))
    return grab_indices(dataset, var, indices, unsort_indices, max_request, file_name=file_name)

def grab_indices(dataset, var, indices, unsort_indices, max_request,
                 file_name='', default=False):
    if default: return np.array([])
    dimensions = retrieve_dimension_list(dataset,var)
    return indices_utils.retrieve_slice(dataset.variables[var], indices, unsort_indices,
                                        dimensions[0], dimensions[1:], 0, max_request)



#External:
import os
import datetime
from sqlite3 import DatabaseError
import requests
import requests_cache

def create_single_session(cache=None,expire_after=datetime.timedelta(hours=1),
                          openid=None,username=None,password=None,use_certificates=None):
    #credentials openid,username and password are accepted only for compatibility
    #purposes
    if cache!=None:
        try:
            session=requests_cache.core.CachedSession(cache,expire_after=expire_after)
        except DatabaseError:
            #Corrupted cache:
            try:
                os.remove(cache)
            except:
                pass
            session=requests_cache.core.CachedSession(cache,expire_after=expire_after)
    else:
        #Create a phony in-memory cached session and disable it:
        session=requests.Session()
    return session



def onlineca_get_cert_wget():
    return """
#!/bin/bash
#
# Client script for web service interface to SLCS based on openssl and
# curl
#
# @author P J Kershaw 25/05/2010
#
# @copyright: (C) 2010 STFC
#
# @license: BSD - See top-level LICENCE file for licence details
#
# $Id$
cmdname=$(basename $0)
cmdline_opt=`getopt hU:l:So:c: $*`

usage="Usage: $cmdname [-U Short-Lived Credential Service URI][-l username] ...\n
\n
   Options\n
       -h\t\t\tDisplays usage\n
       -U <uri>\t\tShort-Lived Credential service URI\n
       -l <username>\t\tUsername for the delegated proxy (defaults to \$LOGNAME)\n
       -S\t\t\tpass password from stdin rather prompt from tty\n
       -o <filepath>\t\tOutput location of end entity certificate or delegated proxy (default to stdout)\n
       -c <directory path>\tDirectory containing the trusted CA (Certificate Authority) certificates.  These are used to\n
       \t\t\tverify the identity of the Short-Lived Credential Service.  Defaults to\n 
       \t\t\t${HOME}/.globus/certificates or\n
       \t\t\t/etc/grid-security/certificates if running as root.\n
"


if [ $? != 0 ] ; then
    echo -e $usage >&2 ;
    exit 1 ;
fi

set -- $cmdline_opt

while true ; do
    case "$1" in
        -h) echo -e $usage ; exit 0 ;;
        -U) uri=$2 ; shift 2 ;;
        -l) username=$2 ; shift 2 ;;
        -S) stdin_pass=True ; shift 1 ;;
        -o) outfilepath=$2 ; shift 2 ;;
        -c) cadir=$2 ; shift 2 ;;
        --) shift ; break ;;
        *) echo "Error parsing command line" ; exit 1 ;;
    esac
done

if [ -z $uri ]; then
    echo -e Give the URI for the Short-Lived Credential service logon request;
    echo -e $usage >&2 ;
    exit 1;
fi

# Default to LOGNAME if not set on command line
if [ -z $username ]; then
    username=${LOGNAME}
fi

# Read password
if [ $stdin_pass ]; then
    read password;
else
    stty -echo
    read -p "Enter Short-Lived Credential phrase: " password; echo
    stty echo
fi

# Set-up trust root
if [ -z $cadir ]; then
    if [ ${X509_CERT_DIR} ]; then
        cadir=${X509_CERT_DIR}
    elif [ "$username" = "root" ]; then
        cadir=/etc/grid-security/certificates
    else
        cadir=${HOME}/.globus/certificates
    fi
fi

# Set output file path
if [ -z $outfilepath ]; then
    if [ ${X509_USER_PROXY} ]; then
        outfilepath=${X509_USER_PROXY}
    else
        # Default to stdout
        outfilepath=/dev/stdout
    fi
fi

# Make a temporary file location for the certificate request
certreqfilepath="/tmp/$UID-$RANDOM.csr"

# Generate key pair and request.  The key file is written to the 'key' var
key=$(openssl req -new -newkey rsa:2048 -nodes -keyout /dev/stdout -subj /CN=dummy -out $certreqfilepath 2> /dev/null)

# URL Encode certificate request - allow for '+' symbol in the base64 charset - 
# needs to be hex equivalent

# Post request to Short-Lived Credential service passing username/password for HTTP Basic
# auth based authentication.  
encoded_certreq=$(cat $certreqfilepath|sed s/+/%2B/g)

# Clean up certificate request temporary file
rm -f $certreqfilepath

response=$(wget --secure-protocol TLSv1 --ca-directory=$cadir \
--http-user=$username --http-password=$password \
--post-data "certificate_request=$encoded_certreq" \
-t 1 $uri -O - 2>&1)

# Pull out the response code from the output
wget_statcode_line="awaiting response..."
responsecode=$(echo "$response"|grep "$wget_statcode_line"|awk '{print $6}'|tail -1)
if [ "$responsecode" != "200" ]; then
    echo "Get certificate call failed."
    echo "$response" >&2
    exit 1
fi

# Extract the certificate
cert=$(echo "$response" | openssl x509)
# Simple sanity check on extracted cert
if [[ $cert != -----BEGIN\ CERTIFICATE-----* ]]; then
    echo "Expecting certificate in response; got:"
    echo "$cert" >&2
    exit 1
fi

# Output certificate
echo "$cert" > $outfilepath

# Add key 
echo "$key" >> $outfilepath
"""



#External:
import netCDF4
import h5netcdf.legacyapi as netCDF4_h5
import netcdf4_pydap

import time
import os
import datetime
from socket import error as SocketError
import requests
from urllib2 import HTTPError, URLError
import copy

#Internal:
from . import safe_handling
from .. import netcdf_utils

class queryable_netCDF:
    def __init__(self,file_name,
                 semaphores=dict(),
                 time_var='time',
                 remote_data_node='',
                 cache=None,
                 timeout=120,
                 expire_after=datetime.timedelta(hours=1),
                 session=None,
                 username=None,
                 password=None,
                 authentication_url=None,
                 use_certificates=False):
        self.file_name=file_name
        self.semaphores=semaphores
        self.time_var=time_var

        if (remote_data_node in  self.semaphores.keys()):
            self.semaphore=semaphores[remote_data_node]
            self.handle_safely=True
        else:
            self.semaphore=safe_handling.dummy_semaphore()
            self.handle_safely=False

        self.cache=cache
        self.timeout=timeout
        self.expire_after=expire_after
        self.session=session
        self.authentication_url = authentication_url
        self.username=username
        self.password=password
        self.use_certificates=use_certificates

        if len(self.file_name)>4 and self.file_name[:4]=='http':
            self.use_pydap=True
            self.max_request=450
            #self.use_h5=False
        else:
            self.use_pydap=False
            self.max_request=2048
            try:
                with netCDF4_h5.Dataset(self.file_name,'r') as dataset:
                    pass
                self.use_h5=True
            except:
                self.use_h5=False
        return

    def __enter__(self):
        self.semaphore.acquire()
        return self

    def __exit__(self,type,value,traceback):
        if self.handle_safely:
            #Do not release semaphore right away if data is not local:
            time.sleep(0.01)
        self.semaphore.release()
        return

    def unsafe_handling(self,function_handle,*args,**kwargs):
        #Capture errors. Important to prevent curl errors from being printed:
        if self.use_pydap:
            with netcdf4_pydap.Dataset(self.file_name,cache=self.cache,
                                       timeout=self.timeout,
                                       expire_after=self.expire_after,
                                       session=self.session,
                                       authentication_url=self.authentication_url,
                                       username=self.username,
                                       password=self.password,
                                       use_certificates=self.use_certificates) as dataset:
                output=function_handle(dataset,*args,**kwargs)
        elif self.use_h5:
            with netCDF4_h5.Dataset(self.file_name,'r') as dataset:
                output=function_handle(dataset,*args,**kwargs)
        else:
            with netCDF4.Dataset(self.file_name,'r') as dataset:
                output=function_handle(dataset,*args,**kwargs)
        return output

    def safe_handling(self,function_handle, *args, **kwargs):
        error_statement=' '.join('''
The url {0} could not be opened. 
Copy and paste this url in a browser and try downloading the file.
If it works, you can stop the download and retry using cdb_query. If
it still does not work it is likely that your certificates are either
not available or out of date.'''.splitlines()).format(self.file_name.replace('dodsC','fileServer'))
        if 'num_trials' in kwargs:
            num_trials = kwargs['num_trials']
            del kwargs['num_trials']
        else:
            num_trials = 5
        success = False
        timeout = copy.copy(self.timeout)
        for trial in range(num_trials):
            if not success:
                try:
                    #Capture errors. Important to prevent curl errors from being printed:
                    if self.use_pydap:
                        with netcdf4_pydap.Dataset(self.file_name,
                                                   cache=self.cache,
                                                   timeout=timeout,
                                                   expire_after=self.expire_after,
                                                   session=self.session,
                                                   authentication_url=self.authentication_url,
                                                   username=self.username,
                                                   password=self.password,
                                                   use_certificates=self.use_certificates) as dataset:
                            try:
                                output=function_handle(dataset,*args,**kwargs)
                            except EOFError as e:
                                #There is an issue with the remote file. Return default:
                                output=function_handle(dataset,*args,default=True,**kwargs)
                    elif self.use_h5:
                        with netCDF4_h5.Dataset(self.file_name,'r') as dataset:
                            output=function_handle(dataset,*args,**kwargs)
                    else:
                        with netCDF4.Dataset(self.file_name,'r') as dataset:
                            output=function_handle(dataset,*args,**kwargs)
                    success=True
                except RuntimeError:
                    time.sleep(3*(trial+1))
                    pass
                except requests.exceptions.ReadTimeout as e:
                    time.sleep(3*(trial+1))
                    #Increase timeout:
                    timeout+=self.timeout
                    pass
                except HTTPError as e:
                    time.sleep(3*(trial+1))
                    #Increase timeout:
                    timeout+=self.timeout
                    pass
                except URLError as e:
                    if e.message == '<urlopen error [Errno 110] Connection timed out>':
                        time.sleep(3*(trial+1))
                        #Increase timeout:
                        timeout+=self.timeout
                        pass
                    else:
                        raise
                except requests.exceptions.ConnectionError as e:
                    time.sleep(3*(trial+1))
                    pass
                except SocketError as e:
                    #http://stackoverflow.com/questions/20568216/python-handling-socket-error-errno-104-connection-reset-by-peer
                    if e.errno != errno.ECONNRESET:
                        raise
                    time.sleep(3*(trial+1))
                    pass
        if not success:
            raise dodsError(error_statement)
        return output

    def check_if_opens(self,num_trials=5):
        try:
            return self.safe_handling(netcdf_utils.check_if_opens, num_trials=num_trials)
        except dodsError as e:
            return False

    def download(self,var,pointer_var,dimensions=dict(),unsort_dimensions=dict(),sort_table=[],time_var='time'):
        retrieved_data=self.safe_handling(
                         netcdf_utils.retrieve_container,var,
                                                        dimensions,
                                                        unsort_dimensions,
                                                        sort_table,self.max_request,
                                                        time_var=self.time_var,
                                                        file_name=self.file_name
                                        )
        return (retrieved_data, sort_table, pointer_var+[var])

class dodsError(Exception):
    def __init__(self, value):
        self.value = value
    def __str__(self):
        return repr(self.value)



#External:
import netCDF4
import numpy as np
import time
import sys
import os
import datetime
import netcdf4_pydap.cas.esgf as esgf

#Internal:
from . import timeaxis_mod, queryable_netcdf, http_netcdf
from .. import netcdf_utils

local_queryable_file_types=['local_file','soft_links_container']
remote_queryable_file_types=['OPENDAP']
queryable_file_types=local_queryable_file_types+remote_queryable_file_types
downloadable_file_types=['FTPServer','HTTPServer','GridFTP']

class remote_netCDF:
    def __init__(self,filename,file_type,
                 semaphores=dict(),
                 session=None,
                 data_node=[],
                 Xdata_node=[],
                 time_var='time',
                 username=None,
                 password=None,
                 openid=None,
                 use_certificates=False,
                 cache=None,
                 timeout=120,
                 expire_after=datetime.timedelta(hours=1)):
        self.filename=filename
        self.file_type=file_type
        self.remote_data_node=get_data_node(self.filename,self.file_type)
        self.semaphores=semaphores

        self.data_node=data_node
        self.Xdata_node=Xdata_node

        self.cache=cache
        self.timeout=timeout
        self.time_var=time_var
        self.expire_after=expire_after
        self.session=session
        self.authentication_url=esgf.authentication_url(openid)
        self.openid=openid
        self.username=username
        self.password=password
        self.use_certificates=use_certificates
        return
    
    def is_available(self,num_trials=5):
        #Do not cache these responses. Must check EVERY time for robustness:
        if not self.file_type in queryable_file_types: 
            with http_netcdf.http_netCDF(self.filename,
                                                semaphores=self.semaphores,
                                                remote_data_node=self.remote_data_node,
                                                timeout=self.timeout,
                                                session=self.session,
                                                authentication_url=self.authentication_url,
                                                username=self.username,
                                                password=self.password,
                                                use_certificates=self.use_certificates) as remote_data:
                return remote_data.check_if_opens(num_trials=num_trials)
        elif not self.file_type in ['soft_links_container']:
            with queryable_netcdf.queryable_netCDF(self.filename,
                                                semaphores=self.semaphores,
                                                remote_data_node=self.remote_data_node,
                                                timeout=self.timeout,
                                                session=self.session,
                                                authentication_url=self.authentication_url,
                                                username=self.username,
                                                password=self.password,
                                                use_certificates=self.use_certificates) as remote_data:
                return remote_data.check_if_opens(num_trials=num_trials)
        else:
            #Any other case, assume true:
            return True

    def download(self,var,pointer_var,download_kwargs=dict()):
        if self.file_type in queryable_file_types:
            with queryable_netcdf.queryable_netCDF(self.filename,
                                                   semaphores=self.semaphores,
                                                   time_var=self.time_var,
                                                   remote_data_node=get_data_node(self.filename,self.file_type),
                                                   cache=self.cache,timeout=self.timeout,
                                                   expire_after=self.expire_after,session=self.session,
                                                   authentication_url=self.authentication_url,
                                                   username=self.username,
                                                   password=self.password,
                                                   use_certificates=self.use_certificates) as remote_data:
                return remote_data.download(var,pointer_var,**download_kwargs)
        elif self.file_type == 'HTTPServer':
            with http_netcdf.http_netCDF(self.filename,
                                         semaphores=self.semaphores,
                                         remote_data_node=get_data_node(self.filename,self.file_type),
                                         cache=self.cache,timeout=self.timeout,
                                         expire_after=self.expire_after,session=self.session,
                                         authentication_url=self.authentication_url,
                                         username=self.username,
                                         password=self.password,
                                         use_certificates=self.use_certificates) as remote_data:
                return remote_data.download(var,pointer_var,**download_kwargs)
        elif self.file_type == 'FTPServer':
            with ftp_netcdf.ftp_netCDF(self.filename,
                                       semaphores=self.semaphores,
                                       remote_data_node=get_data_node(self.filename,self.file_type),
                                       cache=self.cache,timeout=self.timeout,
                                       expire_after=self.expire_after,session=self.session) as remote_data:
                return remote_data.download(var,pointer_var,**download_kwargs)

    def check_if_available_and_find_alternative(self,paths_list,
                                                    file_type_list,
                                                    checksum_list,
                                                    acceptable_file_types,num_trials=5):
        if ( not self.file_type in acceptable_file_types or
             not self.is_available(num_trials=num_trials)):
            checksum=checksum_list[list(paths_list).index(self.filename)]
            for cs_id, cs in enumerate(checksum_list):
                if ( cs==checksum and 
                     paths_list[cs_id]!=self.filename and
                     file_type_list[cs_id] in acceptable_file_types  and
                     is_level_name_included_and_not_excluded('data_node',self,get_data_node(paths_list[cs_id],file_type_list[cs_id]))
                     ):
                        remote_data=remote_netCDF(paths_list[cs_id],
                                                  file_type_list[cs_id],
                                                  self.semaphores,
                                                  cache=self.cache,
                                                  timeout=self.timeout,
                                                  expire_after=self.expire_after,
                                                  session=self.session,
                                                  openid=self.openid,
                                                  username=self.username,
                                                  password=self.password,
                                                  use_certificates=self.use_certificates)
                        if remote_data.is_available(num_trials=num_trials):
                            return paths_list[cs_id]
            return None
        else:
            return self.filename

    def safe_handling(self,function_handle,*args,**kwargs):
        if self.file_type in queryable_file_types:
            with queryable_netcdf.queryable_netCDF(self.filename,
                                                semaphores=self.semaphores,
                                                remote_data_node=get_data_node(self.filename,self.file_type),
                                                cache=self.cache,timeout=self.timeout,
                                                expire_after=self.expire_after,session=self.session,
                                                authentication_url=self.authentication_url,
                                                username=self.username,
                                                password=self.password,
                                                use_certificates=self.use_certificates) as remote_data:
                return remote_data.safe_handling(function_handle,*args,**kwargs)
        else:
            kwargs['default']=True
            return function_handle(None,*args,**kwargs)

    def get_time(self,time_frequency=None,is_instant=False,time_var='time',calendar='standard'):
        if self.file_type in queryable_file_types:
            with queryable_netcdf.queryable_netCDF(self.filename,
                                                semaphores=self.semaphores,
                                                remote_data_node=get_data_node(self.filename,self.file_type),
                                                cache=self.cache,
                                                timeout=self.timeout,
                                                expire_after=self.expire_after,
                                                session=self.session,
                                                authentication_url=self.authentication_url,
                                                username=self.username,
                                                password=self.password,
                                                use_certificates=self.use_certificates) as remote_data:
                return remote_data.safe_handling(netcdf_utils.get_time,time_var=time_var)
        elif time_frequency!=None:
            start_date, end_date = dates_from_filename(self.filename,calendar)
            units=self.get_time_units(calendar)
            start_id=0

            funits=timeaxis_mod.convert_time_units(units, time_frequency)
            end_id=timeaxis_mod.Date2num(end_date,funits,calendar)

            inc = timeaxis_mod.time_inc(time_frequency)
            length = max(end_id/inc-2,1.0)
            
            last_rebuild = start_date
            #use date2num to safely convert dates. Otherwise it might fail for dates <1900:
            if (netCDF4.date2num(last_rebuild, units, calendar=calendar) == 
                netCDF4.date2num(end_date, units, calendar=calendar)):
                date_axis = rebuild_date_axis(0, length, is_instant, inc, funits, calendar=calendar)
                return date_axis

            #use date2num to safely convert dates. Otherwise it might fail for dates <1900:
            while (netCDF4.date2num(last_rebuild, units, calendar=calendar) <
                   netCDF4.date2num(end_date, units, calendar=calendar)):
                date_axis = rebuild_date_axis(0, length, is_instant, inc, funits, calendar=calendar)
                last_rebuild = date_axis[-1]
                length += 1
            return date_axis
        else:
            raise StandardError('time_frequency not provided for non-queryable file type.')
            return date_axis

    def get_calendar(self,time_var='time'):
        if self.file_type in queryable_file_types:
            with queryable_netcdf.queryable_netCDF(self.filename,
                                                semaphores=self.semaphores,
                                                remote_data_node=get_data_node(self.filename,self.file_type),
                                                cache=self.cache,
                                                timeout=self.timeout,
                                                expire_after=self.expire_after,
                                                session=self.session,
                                                authentication_url=self.authentication_url,
                                                username=self.username,
                                                password=self.password,
                                                use_certificates=self.use_certificates) as remote_data:
                return remote_data.safe_handling(netcdf_utils.netcdf_calendar,time_var=time_var)
        else:
            calendar='standard'
        return calendar

    def get_time_units(self,calendar,time_var='time'):
        #Get units from filename:
        if self.file_type in queryable_file_types:
            with queryable_netcdf.queryable_netCDF(self.filename,
                                                semaphores=self.semaphores,
                                                remote_data_node=get_data_node(self.filename,self.file_type),
                                                cache=self.cache,
                                                timeout=self.timeout,
                                                expire_after=self.expire_after,
                                                session=self.session,
                                                authentication_url=self.authentication_url,
                                                username=self.username,
                                                password=self.password,
                                                use_certificates=self.use_certificates) as remote_data:
                return remote_data.safe_handling(netcdf_utils.netcdf_time_units,time_var=time_var)
        else:
            start_date,end_date=dates_from_filename(self.filename,calendar)
            units='days since '+str(start_date)
        return units

def dates_from_filename(filename, calendar):
    """
    Returns datetime objetcs for start and end dates from the filename.

    :param str filename: The filename
    :param str calendar: The NetCDF calendar attribute

    :returns: ``datetime`` instances for start and end dates from the filename
    :rtype: *datetime.datetime*
    This code is adapted from cmip5-timeaxis.

    """
    dates = []
    for date in filename.split('.')[-2].split('_')[-1].split('-'):
        digits = timeaxis_mod.untroncated_timestamp(date)
        # Convert string digits to %Y-%m-%d %H:%M:%S format
        date_as_since = ''.join([''.join(triple) for triple in zip(digits[::2], digits[1::2], ['', '-', '-', ' ', ':', ':', ':'])])[:-1]
        # Use num2date to create netCDF4 datetime objects
        dates.append(netCDF4.num2date(0.0, units='days since ' + date_as_since, calendar=calendar))
    return dates

def rebuild_date_axis(start, length, instant, inc, units, calendar='standard'):
    """
    Rebuilds date axis from numerical time axis, depending on MIP frequency, calendar and instant status.

    :param float date: The numerical date to start (from ``netCDF4.date2num`` or :func:`Date2num`)
    :param int length: The time axis length (i.e., the timesteps number)
    :param boolean instant: The instant status (from :func:`is_instant_time_axis`)
    :param int inc: The time incrementation (from :func:`time_inc`)

    :returns: The corresponding theoretical date axis
    :rtype: *datetime array*

    """
    num_axis = np.arange(start=start, stop=start + length * inc, step=inc)
    if units.split(' ')[0] in ['years', 'months']:
        last = timeaxis_mod.Num2date(num_axis[-1], units=units, calendar=calendar)[0]
    else:
        last = timeaxis_mod.Num2date(num_axis[-1], units=units, calendar=calendar)
    if not instant and not inc in [3, 6]:  # To solve non-instant [36]hr files
        num_axis += 0.5 * inc
    date_axis = timeaxis_mod.Num2date(num_axis, units=units, calendar=calendar)
    return date_axis

def is_level_name_included_and_not_excluded(level_name,options,group):
    if level_name in dir(options):
        if isinstance(getattr(options,level_name),list):
            included=((getattr(options,level_name)==[]) or
                     (group in getattr(options,level_name)))
        else:
            included=((getattr(options,level_name)==None) or 
                       (getattr(options,level_name)==group)) 
    else:
        included=True

    if 'X'+level_name in dir(options):
        if isinstance(getattr(options,'X'+level_name),list):
            not_excluded=((getattr(options,'X'+level_name)==[]) or
                     (not group in getattr(options,'X'+level_name)))
        else:
            not_excluded=((getattr(options,'X'+level_name)==None) or 
                           (getattr(options,'X'+level_name)!=group)) 
    else:
        not_excluded=True
    return included and not_excluded

def get_data_node(path,file_type):
    if file_type=='HTTPServer':
        return '/'.join(path.split('/')[:3])
    elif file_type=='OPENDAP':
        return '/'.join(path.split('/')[:3])
    elif file_type=='FTPServer':
        return '/'.join(path.split('/')[:3])
    elif file_type=='local_file':
        return '/'.join(path.split('/')[:2])
        #return path.split('/')[0]
    elif file_type=='soft_links_container':
        return 'soft_links_container'
    else:
        return ''
